# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutronclient.common import exceptions as n_exceptions
from oslo_log import log

from kuryr.lib import binding
from kuryr.lib import exceptions
from kuryr.lib import segmentation_type_drivers as seg_driver

from kuryr_libnetwork import app
from kuryr_libnetwork.port_driver import base

LOG = log.getLogger(__name__)


class VlanDriver(base.BaseNestedDriver):
    """Driver for container-in-VM deployments with Trunk Ports."""

    BINDING_DRIVERS = ('vlan',)

    def __init__(self):
        super(VlanDriver, self).__init__()

        self.port_vlan_dic = {}
        self.trunk_port = self._get_port_from_host_iface(self.link_iface)
        self._check_for_vlan_ids()

    def _check_for_vlan_ids(self):
        """Gathers information about vlans already in use."""
        for subport in self.trunk_port['trunk_details']['sub_ports']:
            self.port_vlan_dic[subport['port_id']] = subport['segmentation_id']

    def get_supported_bindings(self):
        """Returns a tuple of supported binding driver names for the driver.

        :returns: a tuple of strings
        """
        return self.BINDING_DRIVERS

    def get_default_network_id(self):
        """Returns a Neutron network ID as per driver logic, if any.

        :returns: the Neutron network UUID as a string
        :raises: exceptions.KuryrException
        """
        return None

    def update_port(self, port, endpoint_id, interface_mac):
        segmentation_id = self._get_segmentation_id(port['id'])
        self._attach_subport(self.trunk_port['trunk_details']['trunk_id'],
                             port['id'],
                             segmentation_id)
        return super(VlanDriver, self).update_port(port, endpoint_id,
                                                   interface_mac)

    def create_host_iface(self, endpoint_id, neutron_port, subnets,
                          network=None):
        """Instantiates a host interface and binds it to the host.

        A host linked interface will be created for the specific Neutron port
        by delegating to the pre-selected kuryr-lib driver.
        This driver will attach the port to the trunk port as a subport by
        using a segmentation id available.

        :param endpoint_id:  the ID of the endpoint as string
        :param neutron_port: the container Neutron port dictionary as returned
                             by python-neutronclient
        :param subnets:      an iterable of all the Neutron subnets which the
                             endpoint is trying to join
        :param network:      the Neutron network which the endpoint is trying
                             to join
        :returns: the tuple of stdout and stderr returned by
                  processutils.execute invoked
                  with the executable script for binding
        :raises: exceptions.VethCreationFailure,
                 exceptions.KuryrException,
                 n_exceptions.NeutronClientException,
                 processutils.ProcessExecutionError
        """
        container_ips = neutron_port['fixed_ips']

        if not container_ips:
            raise exceptions.KuryrException(
                "Neutron port {0} does not have fixed_ips."
                .format(neutron_port['id']))

        vm_port = self._get_port_from_host_iface(self.link_iface)

        segmentation_id = self._get_segmentation_id(neutron_port['id'])

        _, _, (stdout, stderr) = binding.port_bind(
            endpoint_id, neutron_port, subnets, network, vm_port,
            segmentation_id)

        return (stdout, stderr)

    def delete_host_iface(self, endpoint_id, neutron_port):
        """Deletes a host interface after unbinding it from the host.

        The host Slave interface associated to the Neutron port will be deleted
        by delegating to the selected kuryr-lib driver.
        This driver will also remove the subport attached to the trunk port
        and will release its segmentation id

        :param endpoint_id:  the ID of the endpoint as string
        :param neutron_port: a port dictionary returned from
                             python-neutronclient
        :returns: the tuple of stdout and stderr returned
                  by processutils.execute invoked with the executable script
                  for unbinding
        :raises: exceptions.VethDeletionFailure,
                 exceptions.KuryrException,
                 n_exceptions.NeutronClientException,
                 processutils.ProcessExecutionError,
        """
        vm_port = self._get_port_from_host_iface(self.link_iface)

        stdout, stderr = binding.port_unbind(endpoint_id, neutron_port)

        subports = [{'port_id': neutron_port['id']}]
        try:
            app.neutron.trunk_remove_subports(
                vm_port['trunk_details']['trunk_id'],
                {'sub_ports': subports})
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during subport deletion "
                      "%(port_id)s: %(ex)s",
                      {'port_id': neutron_port['id'], 'ex': ex})
            raise
        self._release_segmentation_id(neutron_port['id'])
        return stdout, stderr

    def _attach_subport(self, trunk_id, port_id, segmentation_id):
        subport = [
            {
                'segmentation_id': segmentation_id,
                'port_id': port_id,
                'segmentation_type': 'vlan'
            }
        ]
        try:
            app.neutron.trunk_add_subports(trunk_id, {'sub_ports': subport})
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened adding subport %(port_id)s "
                      "to trunk port %(trunk_id)s: %(ex)s",
                      port_id, trunk_id, ex)
            raise

    def _get_segmentation_id(self, id):
        if id in self.port_vlan_dic.keys():
            return self.port_vlan_dic[id]
        seg_id = seg_driver.allocate_segmentation_id(
            self.port_vlan_dic.values())
        self.port_vlan_dic[id] = seg_id
        return seg_id

    def _release_segmentation_id(self, id):
        seg_driver.release_segmentation_id(id)
        del self.port_vlan_dic[id]

    def _get_port_vlan(self, port_id):
        return self.port_vlan_dic[port_id]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys
import time
import traceback

import flask
import jsonschema

from neutronclient.common import exceptions as n_exceptions
from oslo_concurrency import processutils
from oslo_log import log
from werkzeug import exceptions as w_exceptions

from kuryr.lib import constants as lib_const
from kuryr.lib import exceptions
from kuryr.lib import utils as lib_utils
from kuryr_libnetwork import constants as const


LOG = log.getLogger(__name__)
SG_POSTFIX = 'exposed_ports'


# Return all errors as JSON. From http://flask.pocoo.org/snippets/83/
def make_json_app(import_name, **kwargs):
    """Creates a JSON-oriented Flask app.

    All error responses that you don't specifically manage yourself will have
    application/json content type, and will contain JSON that follows the
    libnetwork remote driver protocol.


    { "Err": "405: Method Not Allowed" }


    See:
      - https://github.com/docker/libnetwork/blob/3c8e06bc0580a2a1b2440fe0792fbfcd43a9feca/docs/remote.md#errors  # noqa
    """
    app = flask.Flask(import_name, **kwargs)

    @app.errorhandler(exceptions.KuryrException)
    @app.errorhandler(n_exceptions.NeutronClientException)
    @app.errorhandler(jsonschema.ValidationError)
    @app.errorhandler(processutils.ProcessExecutionError)
    def make_json_error(ex):
        LOG.error("Unexpected error happened: %s", ex)
        traceback.print_exc(file=sys.stderr)
        response = flask.jsonify({"Err": str(ex)})
        response.status_code = w_exceptions.InternalServerError.code
        if isinstance(ex, w_exceptions.HTTPException):
            response.status_code = ex.code
        elif isinstance(ex, n_exceptions.NeutronClientException):
            response.status_code = ex.status_code
        elif isinstance(ex, jsonschema.ValidationError):
            response.status_code = w_exceptions.BadRequest.code
        content_type = 'application/vnd.docker.plugins.v1+json; charset=utf-8'
        response.headers['Content-Type'] = content_type
        return response

    for code in w_exceptions.default_exceptions:
        app.register_error_handler(code, make_json_error)

    return app


def get_sandbox_key(container_id):
    """Returns a sandbox key constructed with the given container ID.

    :param container_id: the ID of the Docker container as string
    :returns: the constructed sandbox key as string
    """
    return os.path.join(lib_utils.DOCKER_NETNS_BASE, container_id[:12])


def get_neutron_port_name(docker_endpoint_id):
    """Returns a Neutron port name.

    :param docker_endpoint_id: the EndpointID
    :returns: the Neutron port name formatted appropriately
    """
    return '-'.join([docker_endpoint_id, lib_utils.PORT_POSTFIX])


def get_sg_expose_name(port_id):
    """Returns a Neutron security group name.

    :param port_id: The Neutron port id to create a security group for
    :returns: the Neutron security group name formatted appropriately
    """
    return '-'.join([port_id, SG_POSTFIX])


def create_net_tags(tag):
    tags = []
    tags.append(const.NEUTRON_ID_LH_OPTION + ':' + tag[:32])
    if len(tag) > 32:
        tags.append(const.NEUTRON_ID_UH_OPTION + ':' + tag[32:64])

    return tags


def existing_net_tag(netid):
    return const.KURYR_EXISTING_NEUTRON_NET + ':' + netid[:12]


def make_net_tags(tag):
    tags = create_net_tags(tag)
    return ','.join(map(str, tags))


def make_net_name(netid, tags=True):
    if tags:
        return const.NET_NAME_PREFIX + netid[:8]
    return netid


def make_subnet_name(pool_cidr):
    return const.SUBNET_NAME_PREFIX + pool_cidr


def create_port_tags(tag):
    tags = []
    tags.append(const.NEUTRON_ID_LH_OPTION + ':' + tag[:32])
    if len(tag) > 32:
        tags.append(const.NEUTRON_ID_UH_OPTION + ':' + tag[32:64])

    return tags


def make_port_tags(tag):
    tags = create_port_tags(tag)
    return ','.join(map(str, tags))


def wait_for_port_active(neutron_client, neutron_port_id, vif_plug_timeout):
    port_active = False
    tries = 0
    while True:
        try:
            port = neutron_client.show_port(neutron_port_id)
        except n_exceptions.NeutronClientException as ex:
            LOG.error('Could not get the port %s to check '
                      'its status', ex)
        else:
            if port['port']['status'] == lib_const.PORT_STATUS_ACTIVE:
                port_active = True
        if port_active or (vif_plug_timeout > 0 and
                           tries >= vif_plug_timeout):
            break
        LOG.debug('Waiting for port %s to become ACTIVE', neutron_port_id)
        tries += 1
        time.sleep(1)

    return port_active



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from collections import defaultdict
import flask
import ipaddress
from itertools import groupby
import jsonschema
import math
from operator import itemgetter
import time

from neutronclient.common import exceptions as n_exceptions
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import strutils

from kuryr.lib import constants as lib_const
from kuryr.lib import exceptions
from kuryr.lib import utils as lib_utils

from kuryr_libnetwork import app
from kuryr_libnetwork import config
from kuryr_libnetwork import constants as const
from kuryr_libnetwork.port_driver import driver
from kuryr_libnetwork import schemata
from kuryr_libnetwork import utils

LOG = log.getLogger(__name__)

MANDATORY_NEUTRON_EXTENSION = "subnet_allocation"
TAG_NEUTRON_EXTENSION = "tag"
TAG_EXT_NEUTRON_EXTENSION = "tag-ext"
TAG_STANDARD_NEUTRON_EXTENSION = 'standard-attr-tag'
SUBNET_POOLS_V4 = []
SUBNET_POOLS_V6 = []
DEFAULT_DRIVER = driver.get_driver_instance()
try:
    SRIOV_DRIVER = driver.get_driver_instance(name='sriov')
except exceptions.KuryrException:
    SRIOV_DRIVER = None
VNIC_TYPES_DRIVERS_MAPPING = {
    const.VNIC_TYPE_NORMAL: DEFAULT_DRIVER,
    const.VNIC_TYPE_DIRECT: SRIOV_DRIVER,
    const.VNIC_TYPE_MACVTAP: SRIOV_DRIVER,
    const.VNIC_TYPE_DIRECT_PHYSICAL: SRIOV_DRIVER,
}


def get_neutron_client():
    """Creates the Neutron client for communicating with Neutron."""
    return lib_utils.get_neutron_client()


def neutron_client():
    if not hasattr(app, 'neutron'):
        app.neutron = get_neutron_client()
        app.enable_dhcp = cfg.CONF.neutron.enable_dhcp
        app.vif_plug_is_fatal = cfg.CONF.neutron.vif_plugging_is_fatal
        app.vif_plug_timeout = cfg.CONF.neutron.vif_plugging_timeout
        app.neutron.format = 'json'


def check_for_neutron_ext_support():
    """Validates for mandatory extension support availability in neutron."""
    max_attempts = 8
    for attempt in range(max_attempts):
        try:
            app.neutron.show_extension(MANDATORY_NEUTRON_EXTENSION)
            break
        except n_exceptions.NeutronClientException as e:
            if e.status_code == n_exceptions.NotFound.status_code:
                raise exceptions.MandatoryApiMissing(
                    "Neutron extension with alias '{0}' not found"
                                .format(MANDATORY_NEUTRON_EXTENSION))
            elif attempt == max_attempts - 1:
                raise
            else:
                LOG.error("Error happened during retrieving neutron "
                          "extensions")
        except Exception as e:
            if attempt == max_attempts - 1:
                raise
            else:
                LOG.error("Error happened during retrieving neutron "
                          "extensions: %s", e)
        backoff = int(math.pow(2, attempt) - 1)
        time.sleep(backoff)


def check_for_neutron_tag_support():
    """Validates tagging extensions availability in Neutron.

    Checks if either tag, tag-ext or standard-attr-tag is available and sets
    ``app`` properties accordingly.
    """

    app.tag_ext = True
    app.tag = True

    # Check for modern extension first.
    try:
        app.neutron.show_extension(TAG_STANDARD_NEUTRON_EXTENSION)
    except n_exceptions.NeutronClientException:
        pass
    else:
        # Okay, this means we have functionality of both tag and tag_ext,
        # we're good to go!
        return

    # Fallback to legacy extensions.
    for ext in (TAG_NEUTRON_EXTENSION, TAG_EXT_NEUTRON_EXTENSION):
        try:
            app.neutron.show_extension(ext)
        except n_exceptions.NeutronClientException as e:
            ext_param = ext.replace('-', '_')  # identifiers cannot have '-'
            setattr(app, ext_param, False)
            if e.status_code == n_exceptions.NotFound.status_code:
                LOG.warning("Neutron extension %s not supported. "
                            "Continue without using them.", ext)


def load_default_subnet_pools():
    """Load the default subnetpools."""
    global SUBNET_POOLS_V4
    global SUBNET_POOLS_V6
    SUBNET_POOLS_V4 = [cfg.CONF.neutron.default_subnetpool_v4]
    SUBNET_POOLS_V6 = [cfg.CONF.neutron.default_subnetpool_v6]


def load_port_driver():
    app.driver = DEFAULT_DRIVER
    LOG.debug("Using port driver '%s'", str(app.driver))


def get_driver(port):
    vnic_type = port.get('binding:vnic_type', const.VNIC_TYPE_NORMAL)
    driver = VNIC_TYPES_DRIVERS_MAPPING.get(vnic_type)
    if driver is None:
        raise exceptions.KuryrException(
            "No port driver available for VNIC type %s" % vnic_type)
    return driver


def _cache_default_subnetpool_ids(app):
    """Caches IDs of the default subnetpools as app.DEFAULT_POOL_IDS."""
    if not hasattr(app, 'DEFAULT_POOL_IDS'):
        default_subnetpool_id_set = set()
        try:
            subnetpool_names = SUBNET_POOLS_V4 + SUBNET_POOLS_V6
            for subnetpool_name in subnetpool_names:
                subnetpools = app.neutron.list_subnetpools(
                    name=subnetpool_name)
                for subnetpool in subnetpools['subnetpools']:
                    default_subnetpool_id_set.add(subnetpool['id'])
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during retrieving the default"
                      " subnet pools: %s", ex)
        app.DEFAULT_POOL_IDS = frozenset(default_subnetpool_id_set)


def _get_networks_by_attrs(**attrs):
    networks = app.neutron.list_networks(**attrs)
    if len(networks.get('networks', [])) > 1:
        raise exceptions.DuplicatedResourceException(
            "Multiple Neutron networks exist for the params {0}"
            .format(', '.join(['{0}={1}'.format(k, v)
                               for k, v in attrs.items()])))
    return networks['networks']


def _get_subnets_by_attrs(**attrs):
    subnets = app.neutron.list_subnets(**attrs)
    return subnets['subnets']


def _get_ports_by_attrs(**attrs):
    ports = app.neutron.list_ports(**attrs)
    if len(ports.get('ports', [])) > 1:
        raise exceptions.DuplicatedResourceException(
            "Multiple Neutron ports exist for the params {0} "
            .format(', '.join(['{0}={1}'.format(k, v)
                               for k, v in attrs.items()])))
    return ports['ports']


def _get_subnetpools_by_attrs(**attrs):
    subnetpools = app.neutron.list_subnetpools(**attrs)
    if len(subnetpools.get('subnetpools', [])) > 1:
        raise exceptions.DuplicatedResourceException(
            "Multiple Neutron subnetspool exist for the params {0} "
            .format(', '.join(['{0}={1}'.format(k, v)
                               for k, v in attrs.items()])))
    return subnetpools['subnetpools']


def _get_subnets_by_interface_cidr(neutron_network_id,
                                   interface_cidr):
    iface = ipaddress.ip_interface(str(interface_cidr))
    subnets = _get_subnets_by_attrs(
        network_id=neutron_network_id, cidr=str(iface.network))
    if len(subnets) > 2:
        raise exceptions.DuplicatedResourceException(
            "Multiple Neutron subnets exist for the network_id={0}"
            "and cidr={1}"
            .format(neutron_network_id, iface.network))
    return subnets


def _get_subnet_by_name(subnet_name):
    subnets_by_name = _get_subnets_by_attrs(name=subnet_name)
    if not subnets_by_name:
        raise exceptions.NoResourceException(
            "The subnet doesn't exist for the name {0}"
            .format(subnets_by_name))
    elif len(subnets_by_name) > 1:
        raise exceptions.DuplicatedResourceException(
            "Multiple Neutron subnets exist for name {0}"
            .format(subnet_name))
    return subnets_by_name[0]['id']


def _get_neutron_port_from_docker_endpoint(endpoint_id):
    port_name = utils.get_neutron_port_name(endpoint_id)
    filtered_ports = app.neutron.list_ports(name=port_name)
    num_ports = len(filtered_ports.get('ports', []))
    if num_ports == 1:
        return filtered_ports['ports'][0]['id']


def _get_neutron_port_status_from_docker_endpoint(endpoint_id):
    response_port_status = {}
    neutron_port_identifier = _make_port_identifier(endpoint_id)
    filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
    if filtered_ports:
        response_port_status['status'] = filtered_ports[0]['status']
    return response_port_status


def _process_interface_address(port_dict, subnets_dict_by_id,
                               response_interface):
    subnet_id = port_dict['subnet_id']
    subnet = subnets_dict_by_id[subnet_id]
    iface = ipaddress.ip_interface(str(subnet['cidr']))
    address_key = 'Address' if iface.version == 4 else 'AddressIPv6'
    response_interface[address_key] = str(iface)


def _create_port(endpoint_id, neutron_network_id, interface_mac, fixed_ips):
    port = {
        'name': utils.get_neutron_port_name(endpoint_id),
        'admin_state_up': True,
        'network_id': neutron_network_id,
        'device_owner': lib_const.DEVICE_OWNER,
        'device_id': endpoint_id,
        'binding:host_id': lib_utils.get_hostname(),
        'fixed_ips': fixed_ips
    }
    if interface_mac:
        port['mac_address'] = interface_mac
    try:
        rcvd_port = app.neutron.create_port({'port': port})
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during creating a"
                  " Neutron port: %s", ex)
        raise
    return rcvd_port['port']


def _get_fixed_ips_by_interface_cidr(subnets, interface_cidrv4,
                                     interface_cidrv6, fixed_ips):
    for subnet in subnets:
        fixed_ip = [('subnet_id=%s' % subnet['id'])]
        if interface_cidrv4 or interface_cidrv6:
            if subnet['ip_version'] == 4 and interface_cidrv4:
                iface = ipaddress.ip_interface(str(interface_cidrv4))
            elif subnet['ip_version'] == 6 and interface_cidrv6:
                iface = ipaddress.ip_interface(str(interface_cidrv6))
            if str(subnet['cidr']) != str(iface.network):
                continue
            fixed_ip.append('ip_address=%s' % iface.ip)
        fixed_ips.extend(fixed_ip)


def _create_or_update_port(neutron_network_id, endpoint_id,
                           interface_cidrv4, interface_cidrv6, interface_mac):
    subnets = []
    fixed_ips = []
    response_port = []

    subnetsv4 = subnetsv6 = []
    if interface_cidrv4:
        subnetsv4 = _get_subnets_by_interface_cidr(
            neutron_network_id, interface_cidrv4)
    if interface_cidrv6:
        subnetsv6 = _get_subnets_by_interface_cidr(
            neutron_network_id, interface_cidrv6)
    subnets = subnetsv4 + subnetsv6
    if not len(subnets):
        raise exceptions.NoResourceException(
            "No subnet exist for the cidrs {0} and {1} "
            .format(interface_cidrv4, interface_cidrv6))
    if len(subnets) > 2:
        raise exceptions.DuplicatedResourceException(
            "Multiple subnets exist for the cidrs {0} and {1}"
            .format(interface_cidrv4, interface_cidrv6))

    _get_fixed_ips_by_interface_cidr(subnets, interface_cidrv4,
                                     interface_cidrv6, fixed_ips)
    filtered_ports = app.neutron.list_ports(fixed_ips=fixed_ips)
    num_port = len(filtered_ports.get('ports', []))
    if not num_port:
        fixed_ips = (
            lib_utils.get_dict_format_fixed_ips_from_kv_format(fixed_ips))
        response_port = _create_port(endpoint_id, neutron_network_id,
                                     interface_mac, fixed_ips)
    elif num_port == 1:
        port = filtered_ports['ports'][0]
        port_driver = get_driver(port)
        response_port = port_driver.update_port(port, endpoint_id,
                                                interface_mac,
                                                tags=app.tag)
        _neutron_port_add_tags(port, endpoint_id)
    # For the container boot from dual-net, request_address will
    # create two ports(v4 and v6 address), we should only allow one
    # for port bind.
    # There are two cases:
    # 1. User specifies an existing port with v4 address only.
    #    In this case, Kuryr creates the v6 port in ipam_request_address.
    #    We will bind the v4 port and remove the v6 port.
    # 2. Users doesn't specify a port. In this case Kuryr creates
    #    the v4 and v6 ports in ipam_request_address and
    #    we will delete both ports then re-create a dual-port.
    elif num_port == 2:
        response_port = None
        for port in filtered_ports.get('ports', []):
            port_name = port.get('name')
            if str(port_name).startswith(const.KURYR_UNBOUND_PORT):
                app.neutron.delete_port(port['id'])
            else:
                port_driver = get_driver(port)
                response_port = port_driver.update_port(port, endpoint_id,
                                                        interface_mac,
                                                        tags=app.tag)
                _neutron_port_add_tags(port, endpoint_id)
        if not response_port:
            fixed_ips = (
                lib_utils.get_dict_format_fixed_ips_from_kv_format(fixed_ips))
            response_port = _create_port(endpoint_id, neutron_network_id,
                                         interface_mac, fixed_ips)
    else:
        raise exceptions.DuplicatedResourceException(
            "Multiple ports exist for the cidrs {0} and {1}"
            .format(interface_cidrv4, interface_cidrv6))

    return response_port, subnets


def _neutron_net_add_tag(net, tag):
    _neutron_add_tag('networks', net, tag)


def _neutron_net_add_tags(net, tag, tags=True):
    if tags:
        tags = utils.create_net_tags(tag)
        for tag in tags:
            _neutron_net_add_tag(net, tag)


def _neutron_net_remove_tag(net, tag):
    _neutron_remove_tag('networks', net, tag)


def _neutron_net_remove_tags(net, tag):
    tags = utils.create_net_tags(tag)
    for tag in tags:
        _neutron_net_remove_tag(net, tag)


def _neutron_subnetpool_add_tag(pool, tag):
    _neutron_add_tag('subnetpools', pool, tag)


def _neutron_subnetpool_remove_tag(pool, tag):
    _neutron_remove_tag('subnetpools', pool, tag)


def _neutron_subnet_add_tag(subnet, tag):
    _neutron_add_tag('subnets', subnet, tag)


def _neutron_subnet_remove_tag(subnet, tag):
    _neutron_remove_tag('subnets', subnet, tag)


def _neutron_port_add_tag(port, tag):
    _neutron_add_tag('ports', port, tag)


def _neutron_port_add_tags(port, tag):
    if app.tag:
        tags = utils.create_port_tags(tag)
        for tag in tags:
            _neutron_port_add_tag(port, tag)


def _neutron_port_remove_tag(port, tag):
    _neutron_remove_tag('ports', port, tag)


def _neutron_port_remove_tags(port, tag):
    if app.tag:
        tags = utils.create_port_tags(tag)
        for tag in tags:
            _neutron_port_remove_tag(port, tag)


def _neutron_add_tag(resource_type, resource, tag):
    if tag not in resource['tags']:
        try:
            app.neutron.add_tag(resource_type, resource['id'], tag)
        except n_exceptions.NotFound:
            LOG.warning("Neutron tags extension for given "
                        "resource type is not supported, "
                        "cannot add tag to %s.", resource_type)


def _neutron_remove_tag(resource_type, resource, tag):
    if tag in resource['tags']:
        app.neutron.remove_tag(resource_type, resource['id'], tag)


def _make_net_identifier(network_id, tags=True):
    if tags:
        return utils.make_net_tags(network_id)
    return network_id


def _get_networks_by_identifier(identifier):
    if app.tag:
        return _get_networks_by_attrs(tags=identifier)
    return _get_networks_by_attrs(name=identifier)


def _make_port_identifier(endpoint_id):
    if app.tag:
        return utils.make_port_tags(endpoint_id)
    return utils.get_neutron_port_name(endpoint_id)


def _get_ports_by_identifier(identifier):
    if app.tag:
        return _get_ports_by_attrs(tags=identifier)
    return _get_ports_by_attrs(name=identifier)


def _program_expose_ports(options, port_id):
    exposed_ports = options.get(const.DOCKER_EXPOSED_PORTS_OPTION)
    if not exposed_ports:
        return

    sec_group = {
        'name': utils.get_sg_expose_name(port_id),
        'description': 'Docker exposed ports created by Kuryr.'
    }
    try:
        sg = app.neutron.create_security_group({'security_group': sec_group})
        sg_id = sg['security_group']['id']

    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during creating a "
                  "Neutron security group: %s", ex)
        raise exceptions.ExportPortFailure(
            ("Could not create required security group {0} "
             "for setting up exported port ").format(sec_group))

    proto_port_dict = defaultdict(list)
    for exposed in exposed_ports:
        port = exposed['Port']
        proto = exposed['Proto']
        try:
            proto = const.PROTOCOLS[proto]
            proto_port_dict[proto].append(port)
        except KeyError:
            # This should not happen as Docker client catches such errors
            LOG.error("Unrecognizable protocol %s", proto)
            app.neutron.delete_security_group(sg_id)
            raise exceptions.ExportPortFailure(
                ("Bad protocol number for exposed port. Deleting "
                 "the security group {0}.").format(sg_id))

    for proto, port_list in proto_port_dict.items():
        # Sort the port range list
        for key, group in groupby(enumerate(sorted(port_list)),
                                  lambda ix: ix[0] - ix[1]):
            port_range_list = list(map(itemgetter(1), group))

            port_range_min = min(port_range_list)
            port_range_max = max(port_range_list)
            sec_group_rule = {
                'security_group_id': sg_id,
                'direction': 'ingress',
                'port_range_min': port_range_min,
                'port_range_max': port_range_max,
                'protocol': proto
            }

            try:
                app.neutron.create_security_group_rule({'security_group_rule':
                                                        sec_group_rule})
            except n_exceptions.NeutronClientException as ex:
                LOG.error("Error happened during creating a "
                          "Neutron security group "
                          "rule: %s", ex)
                app.neutron.delete_security_group(sg_id)
                raise exceptions.ExportPortFailure(
                    ("Could not create required security group rules {0} "
                     "for setting up exported port ").format(sec_group_rule))

    try:
        sgs = [sg_id]
        port = app.neutron.show_port(port_id)
        port = port.get('port')
        if port:
            existing_sgs = port.get('security_groups')
            if existing_sgs:
                sgs = sgs + existing_sgs

        app.neutron.update_port(port_id,
                                {'port': {'security_groups': sgs}})
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during updating a "
                  "Neutron port: %s", ex)
        app.neutron.delete_security_group(sg_id)
        raise exceptions.ExportPortFailure(
            ("Could not update port with required security groups{0} "
             "for setting up exported port ").format(sgs))


def _get_cidr_from_subnetpool(**kwargs):
    pools = _get_subnetpools_by_attrs(**kwargs)
    if pools:
        pool = pools[0]
        pool_id = pool['id']
        prefixes = pool['prefixes']
        if len(prefixes) > 1:
            LOG.warning("More than one prefixes present. "
                        "Picking first one.")

        return ipaddress.ip_network(str(prefixes[0])), pool_id
    else:
        raise exceptions.NoResourceException(
            "No subnetpools with {0} is found."
            .format(kwargs))


def revoke_expose_ports(port_id):
    sgs = app.neutron.list_security_groups(
        name=utils.get_sg_expose_name(port_id))
    sgs = sgs.get('security_groups')
    if not sgs:
        return
    removing_sgs = [sg['id'] for sg in sgs]

    existing_sgs = []
    port = app.neutron.show_port(port_id)
    port = port.get('port')
    if port:
        existing_sgs = port.get('security_groups')
        for sg in removing_sgs:
            if sg in existing_sgs:
                existing_sgs.remove(sg)
        try:
            app.neutron.update_port(port_id,
                                    {'port':
                                     {'security_groups': existing_sgs}})
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during updating a "
                      "Neutron port with a new list of "
                      "security groups: {0}".format(ex))
    try:
        for sg in removing_sgs:
            app.neutron.delete_security_group(sg)
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during deleting a "
                  "Neutron security group: {0}".format(ex))


def _create_kuryr_subnet(pool_cidr, subnet_cidr, pool_id, network_id, gateway):
    new_kuryr_subnet = [{
        'name': utils.make_subnet_name(pool_cidr),
        'network_id': network_id,
        'ip_version': subnet_cidr.version,
        'cidr': str(subnet_cidr),
        'enable_dhcp': app.enable_dhcp,
    }]
    new_kuryr_subnet[0]['subnetpool_id'] = pool_id
    if gateway:
        new_kuryr_subnet[0]['gateway_ip'] = gateway

    subnets = app.neutron.create_subnet({'subnets': new_kuryr_subnet})
    if app.tag_ext:
        for subnet in subnets['subnets']:
            _neutron_subnet_add_tag(subnet, pool_id)
    LOG.debug("Created kuryr subnet %s", new_kuryr_subnet)


def _create_kuryr_subnetpool(pool_cidr, pool_tag, shared):
    pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)

    kwargs = {'name': pool_name}
    if pool_tag:
        kwargs['tags'] = [pool_tag]
    pools = _get_subnetpools_by_attrs(**kwargs)
    if len(pools):
        raise exceptions.KuryrException(
            "Another pool with same cidr exist. ipam and network"
            " options not used to pass pool name")

    cidr = ipaddress.ip_network(str(pool_cidr))
    new_subnetpool = {
        'name': pool_name,
        'default_prefixlen': cidr.prefixlen,
        'prefixes': [pool_cidr],
        'shared': shared
    }
    LOG.info("Creating subnetpool with the given pool CIDR")
    created_subnetpool_response = app.neutron.create_subnetpool(
        {'subnetpool': new_subnetpool})
    pool = created_subnetpool_response['subnetpool']
    if pool_tag:
        _neutron_subnetpool_add_tag(pool, pool_tag)
    return pool


@app.route('/Plugin.Activate', methods=['POST'])
def plugin_activate():
    """Returns the list of the implemented drivers.

    This function returns the list of the implemented drivers defaults to
    ``[NetworkDriver, IpamDriver]`` in the handshake of the remote driver,
     which happens right before the first request against Kuryr.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#handshake  # noqa
    """
    LOG.debug("Received /Plugin.Activate")
    return flask.jsonify(const.SCHEMA['PLUGIN_ACTIVATE'])


@app.route('/NetworkDriver.GetCapabilities', methods=['POST'])
def plugin_scope():
    """Returns the capability as the remote network driver.

    This function returns the capability of the remote network driver, which is
    ``global`` or ``local`` and defaults to ``local``. With ``global``
    capability, the network information is shared among multipe Docker daemons
    if the distributed store is appropriately configured.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#set-capability  # noqa
    """
    LOG.debug("Received /NetworkDriver.GetCapabilities")
    capabilities = {'Scope': cfg.CONF.capability_scope}
    return flask.jsonify(capabilities)


@app.route('/NetworkDriver.DiscoverNew', methods=['POST'])
def network_driver_discover_new():
    """The callback function for the DiscoverNew notification.

    The DiscoverNew notification includes the type of the
    resource that has been newly discovered and possibly other
    information associated with the resource.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#discovernew-notification  # noqa
    """
    LOG.debug("Received /NetworkDriver.DiscoverNew")
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.DiscoverDelete', methods=['POST'])
def network_driver_discover_delete():
    """The callback function for the DiscoverDelete notification.

    The DiscoverDelete notification includes the type of the
    resource that has been deleted and possibly other
    information associated with the resource.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#discoverdelete-notification  # noqa
    """
    LOG.debug("Received /NetworkDriver.DiscoverDelete")
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.AllocateNetwork', methods=['POST'])
def network_driver_allocate_network():
    """Allocate network specific resources passing network id and network specific config.

    Libnetwork miss this API in their doc.
    https://github.com/docker/libnetwork/issues/1699

    See the following like for more detail about the spec:

      https://github.com/docker/libnetwork/blob/master/driverapi/driverapi.go # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for "
              "/NetworkDriver.AllocateNetwork", json_data)
    # Note(limao): This API will only called in docker swarm mode
    # The returned options are passed to CreateNetwork.
    return flask.jsonify({'Options': json_data.get('Options')})


@app.route('/NetworkDriver.FreeNetwork', methods=['POST'])
def network_driver_free_network():
    """Free network specific resources associated with a given network id.

    Libnetwork miss this API in their doc.
    https://github.com/docker/libnetwork/issues/1699

    See the following like for more detail about the spec:

      https://github.com/docker/libnetwork/blob/master/driverapi/driverapi.go # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for "
              "/NetworkDriver.FreeNetwork", json_data)
    # Note(limao): This API will only called in docker swarm mode,
    #              we do not have network resource to free right now,
    #              so just return SUCCESS.
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.CreateNetwork', methods=['POST'])
def network_driver_create_network():
    """Creates a new Neutron Network which name is the given NetworkID.

    This function takes the following JSON data and delegates the actual
    network creation to the Neutron client. libnetwork's NetworkID is used as
    the name of Network in Neutron. ::

        {
            "NetworkID": string,
            "IPv4Data" : [{
                "AddressSpace": string,
                "Pool": ipv4-cidr-string,
                "Gateway" : ipv4-address,
                "AuxAddresses": {
                    "<identifier1>" : "<ipv4-address1>",
                    "<identifier2>" : "<ipv4-address2>",
                    ...
                }
            }, ...],
            "IPv6Data" : [{
                "AddressSpace": string,
                "Pool": ipv6-cidr-string,
                "Gateway" : ipv6-address,
                "AuxAddresses": {
                    "<identifier1>" : "<ipv6-address1>",
                    "<identifier2>" : "<ipv6-address2>",
                    ...
                }
            }, ...],
            "Options": {
                ...
            }
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#create-network  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.CreateNetwork", json_data)
    jsonschema.validate(json_data, schemata.NETWORK_CREATE_SCHEMA)
    container_net_id = json_data['NetworkID']
    neutron_network_name = utils.make_net_name(container_net_id, tags=app.tag)
    v4_pool_cidr = None
    v6_pool_cidr = None
    v4_gateway_ip = ''
    v6_gateway_ip = ''

    def _get_gateway_ip(ip_data):
        gateway_ip = ''
        if 'Gateway' in ip_data:
            gateway_cidr = ip_data['Gateway']
            gateway_ip = gateway_cidr.split('/')[0]
        return gateway_ip

    if json_data['IPv4Data']:
        v4_pool_cidr = json_data['IPv4Data'][0]['Pool']
        v4_gateway_ip = _get_gateway_ip(json_data['IPv4Data'][0])

    if json_data['IPv6Data']:
        v6_pool_cidr = json_data['IPv6Data'][0]['Pool']
        v6_gateway_ip = _get_gateway_ip(json_data['IPv6Data'][0])

    neutron_uuid = None
    neutron_name = None
    v4_pool_name = ''
    v4_pool_id = ''
    v6_pool_name = ''
    v6_pool_id = ''
    v4_subnet_name = ''
    v4_subnet_id = ''
    v6_subnet_name = ''
    v6_subnet_id = ''
    shared = False
    options = json_data.get('Options')
    if options:
        generic_options = options.get(const.NETWORK_GENERIC_OPTIONS)
        if generic_options:
            v4_subnet_id = generic_options.get(
                const.NEUTRON_SUBNET_UUID_OPTION)
            v4_subnet_name = generic_options.get(
                const.NEUTRON_SUBNET_NAME_OPTION)
            v6_subnet_id = generic_options.get(
                const.NEUTRON_V6_SUBNET_UUID_OPTION)
            v6_subnet_name = generic_options.get(
                const.NEUTRON_V6_SUBNET_NAME_OPTION)
            neutron_uuid = generic_options.get(const.NEUTRON_UUID_OPTION)
            neutron_name = generic_options.get(const.NEUTRON_NAME_OPTION)
            v4_pool_name = generic_options.get(const.NEUTRON_POOL_NAME_OPTION)
            v6_pool_name = generic_options.get(
                const.NEUTRON_V6_POOL_NAME_OPTION)
            v4_pool_id = generic_options.get(
                const.NEUTRON_POOL_UUID_OPTION)
            v6_pool_id = generic_options.get(
                const.NEUTRON_V6_POOL_UUID_OPTION)
            shared = strutils.bool_from_string(generic_options.get(
                const.NEUTRON_SHARED_OPTION, 'False'))

    def _get_pool_id(pool_name, pool_cidr, pool_tags):
        pool_id = ''
        kwargs = {}
        if pool_tags:
            kwargs['tags'] = pool_tags
        if not pool_name and pool_cidr:
            pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)
        if pool_name:
            kwargs['name'] = pool_name
            pools = _get_subnetpools_by_attrs(**kwargs)
            if pools:
                pool_id = pools[0]['id']
            else:
                raise exceptions.KuryrException(
                    ("Specified pool name({0}) does not "
                     "exist.").format(pool_name))
        return pool_id

    def _verify_pool_id(pool_id):
        pools = _get_subnetpools_by_attrs(id=pool_id)
        if not pools:
            raise exceptions.KuryrException(
                ("Specified pool id({0}) does not "
                 "exist.").format(pool_id))
    if v4_subnet_name and not v4_subnet_id:
        v4_subnet_id = _get_subnet_by_name(v4_subnet_name)
    if v4_pool_id:
        _verify_pool_id(v4_pool_id)
    else:
        v4_pool_tags = [v4_subnet_id] if v4_subnet_id else None
        v4_pool_id = _get_pool_id(v4_pool_name, v4_pool_cidr, v4_pool_tags)

    if v6_subnet_name and not v6_subnet_id:
        v6_subnet_id = _get_subnet_by_name(v6_subnet_name)
    if v6_pool_id:
        _verify_pool_id(v6_pool_id)
    else:
        v6_pool_tags = [v6_subnet_id] if v6_subnet_id else None
        v6_pool_id = _get_pool_id(v6_pool_name, v6_pool_cidr, v6_pool_tags)

    # let the user override the driver default
    if not neutron_uuid and not neutron_name:
        try:
            neutron_uuid = app.driver.get_default_network_id()
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Failed to retrieve the default driver "
                      "network due to Neutron error: %s", ex)
            raise

    if not neutron_uuid and not neutron_name:
        network = app.neutron.create_network(
            {'network': {'name': neutron_network_name,
                         "admin_state_up": True,
                         'shared': shared}})['network']
        network_id = network['id']
        _neutron_net_add_tags(network, container_net_id,
                              tags=app.tag)

        LOG.info("Created a new network with name "
                 "%(neutron_network_name)s successfully: %(network)s",
                 {'neutron_network_name': neutron_network_name,
                  'network': network})
    else:
        try:
            if neutron_uuid:
                networks = _get_networks_by_attrs(id=neutron_uuid)
                specified_network = neutron_uuid
            else:
                networks = _get_networks_by_attrs(name=neutron_name)
                specified_network = neutron_name
            if not networks:
                raise exceptions.KuryrException(
                    ("Specified network id/name({0}) does not "
                     "exist.").format(specified_network))
            network = networks[0]
            network_id = network['id']
            network_shared = networks[0]['shared']
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during listing "
                      "Neutron networks: %s", ex)
            raise
        if app.tag:
            _neutron_net_add_tags(network, container_net_id, tags=app.tag)
            _neutron_net_add_tag(network,
                                 utils.existing_net_tag(container_net_id))
        else:
            network = app.neutron.update_network(
                neutron_uuid, {'network': {'name': neutron_network_name}})
            LOG.info("Updated the network with new name "
                     "%(neutron_network_name)s successfully: %(network)s",
                     {'neutron_network_name': neutron_network_name,
                      'network': network})
        if network_shared != shared:
            raise exceptions.ConflictConfigOption(
                'Network %(network_id)s had option '
                'shared=%(network_shared)s, conflict with the given option '
                'shared=%(shared)s', {
                    'network_id': network_id,
                    'network_shared': network_shared,
                    'shared': shared
                })
        LOG.info("Using existing network %s "
                 "successfully", specified_network)

    def _get_existing_neutron_subnets(pool_cidr, network_id):
        cidr = None
        subnets = []
        if pool_cidr:
            cidr = ipaddress.ip_network(str(pool_cidr))
            subnets = _get_subnets_by_attrs(network_id=network_id,
                                            cidr=str(cidr))
        if len(subnets) > 1:
            raise exceptions.DuplicatedResourceException(
                "Multiple Neutron subnets exist for the network_id={0}"
                "and cidr={1}".format(network_id, cidr))
        return cidr, subnets

    v4_cidr, v4_subnets = _get_existing_neutron_subnets(v4_pool_cidr,
                                                        network_id)
    v6_cidr, v6_subnets = _get_existing_neutron_subnets(v6_pool_cidr,
                                                        network_id)

    def _add_tag_for_existing_subnet(subnet, pool_id):
        if len(subnet) == 1:
            _neutron_subnet_add_tag(subnet[0], pool_id)

    # This will add a subnetpool_id(created by kuryr) tag
    # for existing Neutron subnets.
    if app.tag_ext:
        _add_tag_for_existing_subnet(v4_subnets, v4_pool_id)
        _add_tag_for_existing_subnet(v6_subnets, v6_pool_id)

    if not v4_subnets and v4_pool_cidr:
        _create_kuryr_subnet(v4_pool_cidr, v4_cidr, v4_pool_id,
                             network_id, v4_gateway_ip)
    if not v6_subnets and v6_pool_cidr:
        _create_kuryr_subnet(v6_pool_cidr, v6_cidr, v6_pool_id,
                             network_id, v6_gateway_ip)

    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.DeleteNetwork', methods=['POST'])
def network_driver_delete_network():
    """Delete the Neutron Network with name as the given NetworkID.

    This function takes the following JSON data and delegates the actual
    network deletion to the Neutron client. ::

        {
            "NetworkID": string
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#delete-network  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.DeleteNetwork", json_data)
    jsonschema.validate(json_data, schemata.NETWORK_DELETE_SCHEMA)

    container_net_id = json_data['NetworkID']
    neutron_network_identifier = _make_net_identifier(container_net_id,
                                                      tags=app.tag)
    if app.tag:
        existing_network_identifier = neutron_network_identifier + ','
        existing_network_identifier += utils.existing_net_tag(container_net_id)
        try:
            existing_networks = _get_networks_by_identifier(
                existing_network_identifier)
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during listing "
                      "Neutron networks: %s", ex)
            raise

        if existing_networks:
            LOG.warning("Network is a pre existing Neutron "
                        "network, not deleting in Neutron. "
                        "removing tags: %s", existing_network_identifier)
            neutron_net = existing_networks[0]
            neutron_net_id = neutron_net['id']
            _neutron_net_remove_tags(neutron_net, container_net_id)
            _neutron_net_remove_tag(neutron_net,
                                    utils.existing_net_tag(container_net_id))
            # Delete subnets created by kuryr
            filtered_subnets = _get_subnets_by_attrs(
                network_id=neutron_net_id)
            for subnet in filtered_subnets:
                try:
                    subnet_name = subnet.get('name')
                    if str(subnet_name).startswith(const.SUBNET_NAME_PREFIX):
                        app.neutron.delete_subnet(subnet['id'])
                except n_exceptions.Conflict:
                    LOG.error("Subnet %s is in use, "
                              "can't be deleted.", subnet['id'])
                except n_exceptions.NeutronClientException as ex:
                    LOG.error("Error happened during deleting a "
                              "subnet created by kuryr: %s", ex)
            return flask.jsonify(const.SCHEMA['SUCCESS'])

    try:
        filtered_networks = _get_networks_by_identifier(
            neutron_network_identifier)
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during listing "
                  "Neutron networks: %s", ex)
        raise

    if not filtered_networks:
        LOG.warning("Network with identifier %s cannot be found",
                    neutron_network_identifier)
    else:
        neutron_network_id = filtered_networks[0]['id']
        filtered_subnets = _get_subnets_by_attrs(
            network_id=neutron_network_id)
        if len(filtered_subnets) > 2:  # subnets for IPv4 and/or IPv6
            raise exceptions.DuplicatedResourceException(
                "Multiple Neutron subnets exist for the network_id={0} "
                .format(neutron_network_id))
        for subnet in filtered_subnets:
            try:
                subnetpool_id = subnet.get('subnetpool_id', None)

                _cache_default_subnetpool_ids(app)

                if subnetpool_id not in app.DEFAULT_POOL_IDS:
                    # If the subnet to be deleted has any port, when some ports
                    # are referring to the subnets in other words,
                    # delete_subnet throws an exception, SubnetInUse that
                    # extends Conflict. This can happen when the multiple
                    # Docker endpoints are created with the same subnet CIDR
                    # and it's totally the normal case. So we'd just log that
                    # and continue to proceed.
                    app.neutron.delete_subnet(subnet['id'])
            except n_exceptions.Conflict:
                LOG.error("Subnet, %s, is in use. Network can't "
                          "be deleted.", subnet['id'])
                raise
            except n_exceptions.NeutronClientException as ex:
                LOG.error("Error happened during deleting a "
                          "Neutron subnets: %s", ex)
                raise

        try:
            app.neutron.delete_network(neutron_network_id)
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during deleting a "
                      "Neutron network: %s", ex)
            raise
        LOG.info("Deleted the network with ID %s "
                 "successfully", neutron_network_id)
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.CreateEndpoint', methods=['POST'])
def network_driver_create_endpoint():
    """Creates new Neutron Subnets and a Port with the given EndpointID.

    This function takes the following JSON data and delegates the actual
    endpoint creation to the Neutron client mapping it into Subnet and Port. ::

        {
            "NetworkID": string,
            "EndpointID": string,
            "Options": {
                ...
            },
            "Interface": {
                "Address": string,
                "AddressIPv6": string,
                "MacAddress": string
            }
        }

    Then the following JSON response is returned. ::

        {
            "Interface": {
                "Address": string,
                "AddressIPv6": string,
                "MacAddress": string
            }
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#create-endpoint  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for "
              "/NetworkDriver.CreateEndpoint", json_data)
    jsonschema.validate(json_data, schemata.ENDPOINT_CREATE_SCHEMA)

    endpoint_id = json_data['EndpointID']
    neutron_network_identifier = _make_net_identifier(json_data['NetworkID'],
                                                      tags=app.tag)
    filtered_networks = _get_networks_by_identifier(neutron_network_identifier)

    if not filtered_networks:
        return flask.jsonify({
            'Err': "Neutron net associated with identifier {0} doesn't exist."
            .format(neutron_network_identifier)
        })
    else:
        neutron_network_id = filtered_networks[0]['id']
        interface = json_data['Interface'] or {}  # Workaround for null
        interface_cidrv4 = interface.get('Address', '')
        interface_cidrv6 = interface.get('AddressIPv6', '')
        interface_mac = interface.get('MacAddress', '')
        if not interface_cidrv4 and not interface_cidrv6:
            return flask.jsonify({
                'Err': "Interface address v4 or v6 not provided."
            })
        neutron_port, subnets = _create_or_update_port(
            neutron_network_id, endpoint_id, interface_cidrv4,
            interface_cidrv6, interface_mac)
        try:
            port_driver = get_driver(neutron_port)
            (stdout, stderr) = port_driver.create_host_iface(
                endpoint_id, neutron_port, subnets, filtered_networks[0])
            LOG.debug(stdout)
            if stderr:
                LOG.error(stderr)
        except (exceptions.VethCreationFailure,
                exceptions.BindingNotSupportedFailure) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('Preparing the veth '
                          'pair was failed: %s.', ex)
        except processutils.ProcessExecutionError:
            with excutils.save_and_reraise_exception():
                LOG.error('Could not bind the Neutron port to '
                          'the veth endpoint.')
        except (exceptions.KuryrException,
                n_exceptions.NeutronClientException) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('Failed to set up the interface: %s', ex)

        if app.vif_plug_is_fatal:
            port_active = utils.wait_for_port_active(
                app.neutron, neutron_port['id'], app.vif_plug_timeout)
            if not port_active:
                neutron_port_name = neutron_port['name']
                raise exceptions.InactiveResourceException(
                    "Neutron port {0} did not become active on time."
                    .format(neutron_port_name))

        response_interface = {}
        created_fixed_ips = neutron_port['fixed_ips']
        subnets_dict_by_id = {subnet['id']: subnet
                              for subnet in subnets}

        if not interface_mac:
            response_interface['MacAddress'] = neutron_port['mac_address']

        vnic_type = neutron_port.get('binding:vnic_type')
        if vnic_type in const.VNIC_TYPES_SRIOV:
            response_interface.pop('MacAddress', None)

        if not (interface_cidrv4 or interface_cidrv6):
            if 'ip_address' in neutron_port:
                _process_interface_address(
                    neutron_port, subnets_dict_by_id, response_interface)
            for fixed_ip in created_fixed_ips:
                _process_interface_address(
                    fixed_ip, subnets_dict_by_id, response_interface)

        LOG.debug("Response JSON data %s for /NetworkDriver.CreateEndpoint",
                  {'Interface': response_interface})
        return flask.jsonify({'Interface': response_interface})


@app.route('/NetworkDriver.EndpointOperInfo', methods=['POST'])
def network_driver_endpoint_operational_info():
    """Return Neutron Port status with the given EndpointID.

    This function takes the following JSON data and delegates the actual
    endpoint query to the Neutron client mapping it into Port status. ::

        {
            "NetworkID": string,
            "EndpointID": string
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#endpoint-operational-info  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for "
              "/NetworkDriver.EndpointOperInfo", json_data)
    jsonschema.validate(json_data, schemata.ENDPOINT_INFO_SCHEMA)

    endpoint_id = json_data['EndpointID']
    response_port_status = (
        _get_neutron_port_status_from_docker_endpoint(endpoint_id))

    LOG.debug("Response JSON data %s for /NetworkDriver.EndpointOperInfo",
              {'Value': response_port_status})
    return flask.jsonify({'Value': response_port_status})


@app.route('/NetworkDriver.DeleteEndpoint', methods=['POST'])
def network_driver_delete_endpoint():
    """Deletes Neutron Subnets and a Port with the given EndpointID.

    This function takes the following JSON data and delegates the actual
    endpoint deletion to the Neutron client mapping it into Subnet and Port. ::

        {
            "NetworkID": string,
            "EndpointID": string
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#delete-endpoint  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.DeleteEndpoint", json_data)
    jsonschema.validate(json_data, schemata.ENDPOINT_DELETE_SCHEMA)

    neutron_network_identifier = _make_net_identifier(json_data['NetworkID'],
                                                      tags=app.tag)
    endpoint_id = json_data['EndpointID']
    filtered_networks = _get_networks_by_identifier(neutron_network_identifier)

    if not filtered_networks:
        return flask.jsonify({
            'Err': "Neutron net associated with identifier {0} doesn't exit."
            .format(neutron_network_identifier)
        })
    else:
        neutron_port_identifier = _make_port_identifier(endpoint_id)
        filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
        if not filtered_ports:
            raise exceptions.NoResourceException(
                "The port doesn't exist for the identifier {0}"
                .format(neutron_port_identifier))
        neutron_port = filtered_ports[0]

        try:
            port_driver = get_driver(neutron_port)
            stdout, stderr = port_driver.delete_host_iface(
                endpoint_id, neutron_port)
            LOG.debug(stdout)
            if stderr:
                LOG.error(stderr)
        except processutils.ProcessExecutionError:
            with excutils.save_and_reraise_exception():
                LOG.error('Could not unbind the Neutron port from'
                          'the veth endpoint.')
        except exceptions.VethDeletionFailure:
            with excutils.save_and_reraise_exception():
                LOG.error('Cleaning the veth pair up was failed.')
        except (exceptions.KuryrException,
                n_exceptions.NeutronClientException) as ex:
            with excutils.save_and_reraise_exception():
                LOG.error('Error while removing the interface: %s', ex)

        _neutron_port_remove_tags(neutron_port, endpoint_id)

    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.Join', methods=['POST'])
def network_driver_join():
    """Binds a Neutron Port to a network interface attached to a container.

    This function takes the following JSON data, creates a veth pair, put one
    end inside of the container and binds another end to the Neutron Port
    specified in the request. ::

        {
            "NetworkID": string,
            "EndpointID": string,
            "SandboxKey": string,
            "Options": {
                ...
            }
        }

    If the binding is succeeded, the following JSON response is returned.::

        {
            "InterfaceName": {
                SrcName: string,
                DstPrefix: string
            },
            "Gateway": string,
            "GatewayIPv6": string,
            "StaticRoutes": [{
                "Destination": string,
                "RouteType": int,
                "NextHop": string,
            }, ...]
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/remote.md#join  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for /NetworkDriver.Join", json_data)
    jsonschema.validate(json_data, schemata.JOIN_SCHEMA)

    neutron_network_identifier = _make_net_identifier(json_data['NetworkID'],
                                                      tags=app.tag)
    endpoint_id = json_data['EndpointID']
    filtered_networks = _get_networks_by_identifier(neutron_network_identifier)

    if not filtered_networks:
        return flask.jsonify({
            'Err': "Neutron net associated with identifier {0} doesn't exit."
            .format(neutron_network_identifier)
        })
    else:
        neutron_network_id = filtered_networks[0]['id']

        neutron_port_identifier = _make_port_identifier(endpoint_id)
        filtered_ports = _get_ports_by_identifier(neutron_port_identifier)
        if not filtered_ports:
            raise exceptions.NoResourceException(
                "The port doesn't exist for the identifier {0}"
                .format(neutron_port_identifier))
        neutron_port = filtered_ports[0]
        all_subnets = _get_subnets_by_attrs(network_id=neutron_network_id)
        kuryr_subnets = []
        for subnet in all_subnets:
            subnet_name = subnet.get('name')
            if str(subnet_name).startswith(const.SUBNET_NAME_PREFIX):
                kuryr_subnets.append(subnet)
        if len(kuryr_subnets) > 2:
            raise exceptions.DuplicatedResourceException(
                "Multiple Kuryr subnets exist for the network_id={0} "
                .format(neutron_network_id))

        port_driver = get_driver(neutron_port)
        iface_name = port_driver.get_container_iface_name(neutron_port)

        join_response = {
            "InterfaceName": {
                "SrcName": iface_name,
                "DstPrefix": config.CONF.binding.veth_dst_prefix
            },
            "StaticRoutes": [],
            "DisableGatewayService": True,
        }

        for subnet in all_subnets:
            for fixed_ip in neutron_port['fixed_ips']:
                if fixed_ip['subnet_id'] == subnet['id']:
                    if subnet['ip_version'] == 4:
                        join_response['Gateway'] = subnet.get('gateway_ip', '')
                    else:
                        join_response['GatewayIPv6'] = subnet.get(
                            'gateway_ip', '')

            # NOTE: kuryr-libnetwork do not support a connected route
            host_routes = subnet.get('host_routes', [])
            for host_route in host_routes:
                static_route = {
                    'Destination': host_route['destination'],
                    'RouteType': const.ROUTE_TYPE['NEXTHOP'],
                    'NextHop': host_route['nexthop']
                }
                join_response['StaticRoutes'].append(static_route)

        LOG.debug("Response JSON data %s for /NetworkDriver.Join",
                  join_response)
        return flask.jsonify(join_response)


@app.route('/NetworkDriver.Leave', methods=['POST'])
def network_driver_leave():
    """Unbinds a Neutron Port to a network interface attached to a container.

    This function takes the following JSON data and delete the veth pair
    corresponding to the given info. ::

        {
            "NetworkID": string,
            "EndpointID": string
        }
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.Leave", json_data)
    jsonschema.validate(json_data, schemata.LEAVE_SCHEMA)

    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.ProgramExternalConnectivity', methods=['POST'])
def network_driver_program_external_connectivity():
    """Provides external connectivity for a given container.

    Performs the necessary programming to allow the external connectivity
    dictated by the specified options

    See the following link for more details about the spec:
      https://github.com/docker/libnetwork/blob/master/driverapi/driverapi.go
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.ProgramExternalConnectivity", json_data)
    if not cfg.CONF.process_external_connectivity:
        return flask.jsonify(const.SCHEMA['SUCCESS'])

    # TODO(banix): Add support for exposed ports
    port = _get_neutron_port_from_docker_endpoint(json_data['EndpointID'])
    if port:
        _program_expose_ports(json_data['Options'], port)

    # TODO(banix): Add support for published ports
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/NetworkDriver.RevokeExternalConnectivity', methods=['POST'])
def network_driver_revoke_external_connectivity():
    """Removes external connectivity for a given container.

    Performs the necessary programming to remove the external connectivity
    of a container

    See the following link for more details about the spec:
      https://github.com/docker/libnetwork/blob/master/driverapi/driverapi.go
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for"
              " /NetworkDriver.RevokeExternalConnectivity", json_data)
    if not cfg.CONF.process_external_connectivity:
        return flask.jsonify(const.SCHEMA['SUCCESS'])

    # TODO(banix): Add support for removal of exposed ports
    port = _get_neutron_port_from_docker_endpoint(json_data['EndpointID'])
    if port:
        revoke_expose_ports(port)

    # TODO(banix): Add support for removal of published ports
    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/IpamDriver.GetCapabilities', methods=['POST'])
def ipam_get_capabilities():
    """Provides the IPAM driver capabilities.

    This function is called during the registration of the IPAM driver.

    See the following link for more details about the spec:
      https://github.com/docker/libnetwork/blob/master/docs/ipam.md#getcapabilities  # noqa
    """
    LOG.debug("Received /IpamDriver.GetCapabilities")
    capabilities = {'RequiresMACAddress': True}
    return flask.jsonify(capabilities)


@app.route('/IpamDriver.GetDefaultAddressSpaces', methods=['POST'])
def ipam_get_default_address_spaces():
    """Provides the default address spaces for the IPAM.

    This function is called after the registration of the IPAM driver and
    the plugin set the returned values as the default address spaces for the
    IPAM. The address spaces can be configured in the config file.

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/ipam.md#getdefaultaddressspaces  # noqa
    """
    LOG.debug("Received /IpamDriver.GetDefaultAddressSpaces")
    address_spaces = {
        'LocalDefaultAddressSpace': cfg.CONF.local_default_address_space,
        'GlobalDefaultAddressSpace': cfg.CONF.global_default_address_space}
    return flask.jsonify(address_spaces)


@app.route('/IpamDriver.RequestPool', methods=['POST'])
def ipam_request_pool():
    """Creates a new Neutron subnetpool from the given request.

    This funciton takes the following JSON data and delegates the subnetpool
    creation to the Neutron client. ::

        {
            "AddressSpace": string
            "Pool":         string
            "SubPool":      string
            "Options":      map[string]string
            "V6":           bool
        }

    Then the following JSON response is returned. ::

        {
            "PoolID": string
            "Pool":   string
            "Data":   map[string]string
        }

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/ipam.md#requestpool  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for /IpamDriver.RequestPool",
              json_data)
    jsonschema.validate(json_data, schemata.REQUEST_POOL_SCHEMA)
    requested_pool = json_data['Pool']
    requested_subpool = json_data['SubPool']
    v6 = json_data['V6']
    subnet_cidr = ''
    pool_name = ''
    pool_id = ''
    subnet_id = ''
    subnet_name = ''
    shared = False
    options = json_data.get('Options')
    if options:
        shared = strutils.bool_from_string(options.get(
            const.NEUTRON_SHARED_OPTION, 'False'))
        if v6:
            subnet_name = options.get(const.NEUTRON_V6_SUBNET_NAME_OPTION)
            subnet_id = options.get(const.NEUTRON_V6_SUBNET_UUID_OPTION)
            pool_name = options.get(const.NEUTRON_V6_POOL_NAME_OPTION)
            pool_id = options.get(const.NEUTRON_V6_POOL_UUID_OPTION)
        else:
            subnet_name = options.get(const.NEUTRON_SUBNET_NAME_OPTION)
            subnet_id = options.get(const.NEUTRON_SUBNET_UUID_OPTION)
            pool_name = options.get(const.NEUTRON_POOL_NAME_OPTION)
            pool_id = options.get(const.NEUTRON_POOL_UUID_OPTION)
    if requested_pool:
        if requested_subpool:
            cidr = ipaddress.ip_network(str(requested_subpool))
        else:
            cidr = ipaddress.ip_network(str(requested_pool))
        subnet_cidr = str(cidr)
        if not subnet_id and subnet_name:
            subnet_id = _get_subnet_by_name(subnet_name)
        elif not subnet_id and not subnet_name:
            subnets_by_cidr = _get_subnets_by_attrs(cidr=subnet_cidr)
            if len(subnets_by_cidr):
                LOG.warning("There is already existing subnet for the "
                            "same cidr. Please check and specify pool name "
                            "in Options.")
        if not pool_name and not pool_id:
            pool_id = _create_kuryr_subnetpool(subnet_cidr,
                                               subnet_id,
                                               shared)['id']
        else:
            if pool_id:
                existing_pools = _get_subnetpools_by_attrs(id=pool_id)
            else:
                existing_pools = _get_subnetpools_by_attrs(name=pool_name)
            if not existing_pools:
                raise exceptions.KuryrException(
                    ("Specified subnetpool id/name({0}) does not "
                     "exist.").format(pool_id or pool_name))

            pool_id = existing_pools[0]['id']
            if app.tag_ext:
                _neutron_subnetpool_add_tag(
                    existing_pools[0], const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
            prefixes = existing_pools[0]['prefixes']
            pool_cidr = ipaddress.ip_network(str(prefixes[0]))
            if pool_cidr == cidr:
                if shared != existing_pools[0]['shared']:
                    raise exceptions.ConflictConfigOption(
                        'There is already existing subnet pool '
                        'with %(cidr)s but with shared = %(shared)s',
                        {'cidr': cidr,
                         'shared': existing_pools[0]['shared']})
                LOG.info("Using existing Neutron subnetpool %s successfully",
                         pool_id)
            else:
                pool_id = _create_kuryr_subnetpool(subnet_cidr,
                                                   subnet_id,
                                                   shared)['id']
    else:
        if v6:
            default_pool_list = SUBNET_POOLS_V6
        else:
            default_pool_list = SUBNET_POOLS_V4
        pool_name = default_pool_list[0]
        subnet_cidr, pool_id = _get_cidr_from_subnetpool(name=pool_name)
        subnet_cidr = str(subnet_cidr)

    req_pool_res = {'PoolID': pool_id,
                    'Pool': subnet_cidr}
    LOG.debug("Response JSON data %s for /IpamDriver.RequestPool",
              req_pool_res)
    return flask.jsonify(req_pool_res)


@app.route('/IpamDriver.RequestAddress', methods=['POST'])
def ipam_request_address():
    """Allocates the IP address in the given request.

    This function takes the following JSON data and add the given IP address in
    the allocation_pools attribute of the subnet. ::

        {
            "PoolID":  string
            "Address": string
            "Options": map[string]string
        }

    Then the following response is returned. ::

        {
            "Address": string
            "Data":    map[string]string
        }

    See the following link for more details about the spec:

    https://github.com/docker/libnetwork/blob/master/docs/ipam.md#requestaddress  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for "
              "/IpamDriver.RequestAddress", json_data)
    jsonschema.validate(json_data, schemata.REQUEST_ADDRESS_SCHEMA)
    pool_id = json_data['PoolID']
    req_address = json_data['Address']
    req_mac_address = ''
    is_gateway = False
    allocated_address = ''
    subnet = {}
    # Check if the port is gateway
    options = json_data.get('Options')
    if options:
        req_mac_address = options.get(const.DOCKER_MAC_ADDRESS_OPTION)
        request_address_type = options.get(const.REQUEST_ADDRESS_TYPE)
        if request_address_type == const.NETWORK_GATEWAY_OPTIONS:
            is_gateway = True

    # check if any subnet with matching subnetpool_id is present
    subnets_by_poolid = _get_subnets_by_attrs(subnetpool_id=pool_id)
    if subnets_by_poolid:
        if len(subnets_by_poolid) == 1:
            subnet = subnets_by_poolid[0]
            subnet_cidr = ipaddress.ip_network(str(subnet['cidr']))
        else:
            pool_cidr, _ = _get_cidr_from_subnetpool(id=pool_id)
            for tmp_subnet in subnets_by_poolid:
                subnet_cidr = ipaddress.ip_network(
                    str(tmp_subnet['cidr']))
                if pool_cidr == subnet_cidr:
                    subnet = tmp_subnet
                    break
    else:
        # check if any subnet with matching cidr is present
        subnet_cidr, _ = _get_cidr_from_subnetpool(id=pool_id)
        subnets_by_cidr = _get_subnets_by_attrs(
            cidr=str(subnet_cidr))
        if len(subnets_by_cidr) > 1:
            for tmp_subnet in subnets_by_cidr:
                if tmp_subnet.get('tags') is not None:
                    if pool_id in tmp_subnet.get('tags'):
                        subnet = tmp_subnet
                else:
                    LOG.warning("subnetpool tag for Neutron "
                                "subnet %s is missing, cannot "
                                "gets the correct subnet.",
                                tmp_subnet['id'])
        elif len(subnets_by_cidr) == 1:
            subnet = subnets_by_cidr[0]

    if not any(subnet) and not is_gateway:
        raise exceptions.KuryrException(
            ("Subnet with pool {0} does not exist.").format(pool_id))

    if any(subnet):
        if is_gateway:
            # check if request gateway ip same with existed gateway ip
            existed_gateway_ip = subnet.get('gateway_ip', '')
            if not req_address:
                if subnet['ip_version'] == 4:
                    allocated_address = "0.0.0.0/0"
                else:
                    allocated_address = "::/0"
            elif req_address == existed_gateway_ip:
                allocated_address = '{}/{}'.format(req_address,
                                                   subnet_cidr.prefixlen)
            else:
                raise exceptions.GatewayConflictFailure(
                    "Requested gateway {0} does not match with "
                    "gateway {1} in existed "
                    "network.".format(req_address, existed_gateway_ip))
        else:
            # allocating address for container port
            neutron_network_id = subnet['network_id']
            try:
                port = {
                    'name': const.KURYR_UNBOUND_PORT,
                    'admin_state_up': True,
                    'network_id': neutron_network_id,
                }
                fixed_ips = port['fixed_ips'] = []
                fixed_ip = {'subnet_id': subnet['id']}
                filtered_ports = []
                if req_address:
                    fixed_ip['ip_address'] = req_address
                    fixed_ip_existing = [('subnet_id=%s' % subnet['id'])]
                    fixed_ip_existing.append('ip_address='
                                             '%s' % str(req_address))
                    filtered_ports = app.neutron.list_ports(
                        fixed_ips=fixed_ip_existing).get('ports', [])
                if not filtered_ports:
                    filtered_ports = app.neutron.list_ports(
                        fixed_ips='subnet_id=%s' % subnet['id'],
                        mac_address=req_mac_address).get('ports', [])

                num_ports = len(filtered_ports)
                fixed_ips.append(fixed_ip)

                if num_ports:
                    created_port = filtered_ports[0]
                    # REVISIT(yedongcan) For tag-ext extension not
                    # supported, the Neutron existing port still can not
                    # be deleted in ipam_release_address.
                    if app.tag_ext:
                        _neutron_port_add_tag(
                            created_port,
                            const.KURYR_EXISTING_NEUTRON_PORT)
                else:
                    created_port_resp = app.neutron.create_port({'port': port})
                    created_port = created_port_resp['port']
                    if app.tag_ext:
                        _neutron_port_add_tag(created_port,
                                              lib_const.DEVICE_OWNER)

                LOG.debug("created port %s", created_port)
                fixed_ips = created_port['fixed_ips']
                fixed_ips = [ip for ip in fixed_ips
                             if ip['subnet_id'] == subnet['id']]
                allocated_address = (req_address or
                                     fixed_ips[0]['ip_address'])
                allocated_address = '{}/{}'.format(allocated_address,
                                                   subnet_cidr.prefixlen)
            except n_exceptions.NeutronClientException as ex:
                LOG.error("Error happened during ip allocation on "
                          "Neutron side: %s", ex)
                raise
    else:
        # Auxiliary address or gw_address is received at network creation time.
        # This address cannot be reserved with neutron at this time as subnet
        # is not created yet. In /NetworkDriver.CreateNetwork this address will
        # be reserved with neutron.
        if req_address:
            allocated_address = '{}/{}'.format(req_address,
                                               subnet_cidr.prefixlen)

    LOG.debug("Response JSON data %s for /IpamDriver.RequestAddress",
              {'Address': allocated_address})
    return flask.jsonify({'Address': allocated_address})


@app.route('/IpamDriver.ReleasePool', methods=['POST'])
def ipam_release_pool():
    """Deletes a new Neutron subnetpool from the given reuest.

    This function takes the following JSON data and delegates the subnetpool
    deletion to the Neutron client. ::

       {
           "PoolID": string
       }

    Then the following JSON response is returned. ::

       {}

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/ipam.md#releasepool  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for /IpamDriver.ReleasePool",
              json_data)
    jsonschema.validate(json_data, schemata.RELEASE_POOL_SCHEMA)
    pool_id = json_data['PoolID']

    # Remove subnetpool_id tag from Neutron existing subnet.
    if app.tag_ext:
        subnet_cidr, _ = _get_cidr_from_subnetpool(id=pool_id)
        subnets_by_cidr = _get_subnets_by_attrs(
            cidr=str(subnet_cidr))
        for tmp_subnet in subnets_by_cidr:
            if pool_id in tmp_subnet.get('tags', []):
                _neutron_subnet_remove_tag(tmp_subnet, pool_id)
                break

    pools = _get_subnetpools_by_attrs(id=pool_id)
    if pools:
        pool_name = pools[0]['name']
        if app.tag_ext:
            tags = pools[0].get('tags', [])
            if const.KURYR_EXISTING_NEUTRON_SUBNETPOOL in tags:
                _neutron_subnetpool_remove_tag(
                    pools[0], const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
                LOG.debug('Skip the cleanup since this is an existing Neutron '
                          'subnetpool.')
                return flask.jsonify(const.SCHEMA['SUCCESS'])
        elif not pool_name.startswith(cfg.CONF.subnetpool_name_prefix):
            LOG.debug('Skip the cleanup since this is an existing Neutron '
                      'subnetpool.')
            return flask.jsonify(const.SCHEMA['SUCCESS'])

    # Delete subnets created by kuryr
    filtered_subnets = _get_subnets_by_attrs(
        subnetpool_id=pool_id)
    for subnet in filtered_subnets:
        try:
            subnet_name = subnet.get('name')
            if str(subnet_name).startswith(const.SUBNET_NAME_PREFIX):
                app.neutron.delete_subnet(subnet['id'])
        except n_exceptions.Conflict:
            LOG.error("Subnet %s is in use, "
                      "can't be deleted.", subnet['id'])
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during deleting a "
                      "subnet created by kuryr: %s", ex)

    try:
        app.neutron.delete_subnetpool(pool_id)
    except n_exceptions.Conflict:
        LOG.info("The subnetpool with ID %s is still in use."
                 " It can't be deleted for now.", pool_id)
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened during deleting a "
                  "Neutron subnetpool: %s", ex)
        raise

    return flask.jsonify(const.SCHEMA['SUCCESS'])


@app.route('/IpamDriver.ReleaseAddress', methods=['POST'])
def ipam_release_address():
    """Deallocates the IP address in the given request.

    This function takes the following JSON data and remove the given IP address
    from the allocation_pool attribute of the subnet. ::

        {
            "PoolID": string
            "Address": string
        }

    Then the following response is returned. ::

        {}

    See the following link for more details about the spec:

      https://github.com/docker/libnetwork/blob/master/docs/ipam.md#releaseaddress  # noqa
    """
    json_data = flask.request.get_json(force=True)
    LOG.debug("Received JSON data %s for /IpamDriver.ReleaseAddress",
              json_data)
    jsonschema.validate(json_data, schemata.RELEASE_ADDRESS_SCHEMA)
    pool_id = json_data['PoolID']
    rel_address = json_data['Address']
    # check if any subnet with matching subnetpool_id is present
    subnets = _get_subnets_by_attrs(subnetpool_id=pool_id)
    if not len(subnets):
        # check if any subnet with matching cidr is present
        subnet_cidr = str(_get_cidr_from_subnetpool(id=pool_id)[0])
        subnets = _get_subnets_by_attrs(cidr=subnet_cidr)
    if not len(subnets):
        LOG.info("Subnet already deleted.")
        return flask.jsonify(const.SCHEMA['SUCCESS'])
    if app.tag_ext:
        subnets = [subnet
                   for subnet in subnets
                   if pool_id in subnet.get('tags') or []]

    iface = ipaddress.ip_interface(str(rel_address))
    rel_ip_address = str(iface.ip)
    try:
        fixed_ip = 'ip_address=' + str(rel_ip_address)
        all_ports = app.neutron.list_ports(fixed_ips=fixed_ip)
        for port in all_ports['ports']:
            tags = port.get('tags', [])
            if ((tags and lib_const.DEVICE_OWNER in tags) or
                    (not tags and port['name'] ==
                        utils.get_neutron_port_name(port['device_id']))):
                for tmp_subnet in subnets:
                    if (port['fixed_ips'][0]['subnet_id'] == tmp_subnet['id']):
                        app.neutron.delete_port(port['id'])
            elif tags and const.KURYR_EXISTING_NEUTRON_PORT in tags:
                updated_port = {'device_owner': '',
                                'binding:host_id': ''}
                if port['name'].startswith(port['device_id']):
                    updated_port["device_id"] = ''
                for subnet in subnets:
                    if (port['fixed_ips'][0]['subnet_id'] == subnet['id']):
                        app.neutron.update_port(
                            port['id'], {'port': updated_port})
                        _neutron_port_remove_tag(
                            port, const.KURYR_EXISTING_NEUTRON_PORT)
    except n_exceptions.NeutronClientException as ex:
        LOG.error("Error happened while fetching "
                  "and deleting port, %s", ex)
        raise

    return flask.jsonify(const.SCHEMA['SUCCESS'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from kuryr.lib import utils as lib_utils
from kuryr_libnetwork import constants
from kuryr_libnetwork.tests.fullstack import kuryr_base
from kuryr_libnetwork import utils


class NetworkTest(kuryr_base.KuryrBaseTest):
    """Test Networks operation

    Test networks creation/deletion from docker to Neutron
    """
    def test_create_delete_network_with_kuryr_driver(self):
        """Create and Delete docker network with Kuryr

           This method creates a docker network with Kuryr driver
           and tests it was created in Neutron.
           It then deletes the docker network and tests that it was
           deleted from Neutron.
        """
        fake_ipam = {
            "Driver": "kuryr",
            "Options": {},
            "Config": [
                {
                    "Subnet": "10.0.0.0/16",
                    "IPRange": "10.0.0.0/24",
                    "Gateway": "10.0.0.1"
                }
            ]
        }
        net_name = lib_utils.get_random_string(8)
        res = self.docker_client.create_network(name=net_name, driver='kuryr',
                                                ipam=fake_ipam)
        net_id = res['Id']
        try:
            network = self.neutron_client.list_networks(
                tags=utils.make_net_tags(net_id))
        except Exception as e:
            self.docker_client.remove_network(net_id)
            message = ("Failed to list neutron networks: %s")
            self.fail(message % e.args[0])
        self.assertEqual(1, len(network['networks']))
        self.docker_client.remove_network(net_id)
        network = self.neutron_client.list_networks(
            tags=utils.make_net_tags(net_id))
        self.assertEqual(0, len(network['networks']))

    def test_create_delete_dualstack_network_with_kuryr_driver(self):
        """Create and Delete docker dual-stack network with Kuryr

           This method creates a docker network with Kuryr driver
           and tests it was created in Neutron.
           It then deletes the docker network and tests that it was
           deleted from Neutron.
        """
        fake_ipam = {
            "Driver": "kuryr",
            "Options": {},
            "Config": [
                {
                    "Subnet": "10.4.0.0/16",
                    "IPRange": "10.4.0.0/24",
                    "Gateway": "10.4.0.1"
                },
                {
                    "Subnet": "2001:db8:a0b:12f0::/64",
                    "IPRange": "2001:db8:a0b:12f0::/64",
                    "Gateway": "2001:db8:a0b:12f0::1"
                }
            ]
        }
        net_name = lib_utils.get_random_string(8)
        res = self.docker_client.create_network(name=net_name,
                                                driver='kuryr',
                                                enable_ipv6=True,
                                                ipam=fake_ipam)
        net_id = res['Id']
        try:
            network = self.neutron_client.list_networks(
                tags=utils.make_net_tags(net_id))
        except Exception as e:
            self.docker_client.remove_network(net_id)
            message = ("Failed to list neutron networks: %s")
            self.fail(message % e.args[0])
        self.assertEqual(1, len(network['networks']))
        neutron_netid = network['networks'][0]['id']
        subnets = self.neutron_client.list_subnets(
            network_id=neutron_netid)
        self.assertEqual(2, len(subnets['subnets']))
        subnet_version = []
        subnet_name = []
        for subnet in subnets['subnets']:
            subnet_version.append(subnet['ip_version'])
            subnet_name.append(subnet['name'])
        self.assertIn(4, subnet_version)
        self.assertIn(6, subnet_version)
        self.assertIn(
            constants.SUBNET_NAME_PREFIX + '10.4.0.0/24',
            subnet_name)
        self.assertIn(
            constants.SUBNET_NAME_PREFIX + '2001:db8:a0b:12f0::/64',
            subnet_name)
        self.docker_client.remove_network(net_id)
        network = self.neutron_client.list_networks(
            tags=utils.make_net_tags(net_id))
        self.assertEqual(0, len(network['networks']))
        subnets = self.neutron_client.list_subnets(
            network_id=neutron_netid)
        self.assertEqual(0, len(subnets['subnets']))

    def test_create_delete_network_without_kuryr_driver(self):
        """Create and Delete docker network without Kuryr

           This method create a docker network with the default
           docker driver, It tests that it was created correctly, but
           not added to Neutron
        """
        net_name = lib_utils.get_random_string(8)
        res = self.docker_client.create_network(name=net_name)
        net_id = res['Id']
        network = self.neutron_client.list_networks(
            tags=utils.make_net_tags(net_id))
        self.assertEqual(0, len(network['networks']))
        docker_networks = self.docker_client.networks()
        network_found = False
        for docker_net in docker_networks:
            if docker_net['Id'] == net_id:
                network_found = True
        self.assertTrue(network_found)
        self.docker_client.remove_network(net_id)

    def test_create_network_with_same_name(self):
        """Create docker network with same name

           Create two docker networks with same name,
           delete them and see that neutron networks are
           deleted as well
        """
        fake_ipam_1 = {
            "Driver": "kuryr",
            "Options": {},
            "Config": [
                {
                    "Subnet": "10.1.0.0/16",
                    "IPRange": "10.1.0.0/24",
                    "Gateway": "10.1.0.1"
                }
            ]
        }
        fake_ipam_2 = {
            "Driver": "kuryr",
            "Options": {},
            "Config": [
                {
                    "Subnet": "10.2.0.0/16",
                    "IPRange": "10.2.0.0/24",
                    "Gateway": "10.2.0.1"
                }
            ]
        }
        net_name = lib_utils.get_random_string(8)
        res = self.docker_client.create_network(name=net_name, driver='kuryr',
                                                ipam=fake_ipam_1)
        net_id1 = res['Id']

        res = self.docker_client.create_network(name=net_name, driver='kuryr',
                                                ipam=fake_ipam_2)
        net_id2 = res['Id']
        try:
            network = self.neutron_client.list_networks(
                tags=utils.make_net_tags(net_id1))
            self.assertEqual(1, len(network['networks']))
            network = self.neutron_client.list_networks(
                tags=utils.make_net_tags(net_id2))
            self.assertEqual(1, len(network['networks']))
        except Exception as e:
            self.docker_client.remove_network(net_id1)
            self.docker_client.remove_network(net_id2)
            message = ("Failed to list neutron networks: %s")
            self.fail(message % e.args[0])
        self.docker_client.remove_network(net_id1)
        self.docker_client.remove_network(net_id2)
        network = self.neutron_client.list_networks(
            tags=utils.make_net_tags(net_id1))
        self.assertEqual(0, len(network['networks']))
        network = self.neutron_client.list_networks(
            tags=utils.make_net_tags(net_id2))
        self.assertEqual(0, len(network['networks']))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

EPSILON_PATTERN = '^$'
IPV4_PATTERN_BASE = (u'((25[0-5]|2[0-4][0-9]|1?[0-9]?[0-9])\\.){3}'
                     u'(25[0-5]|2[0-4][0-9]|1?[0-9]?[0-9])')
IPV4_PATTERN = EPSILON_PATTERN + u'|^' + IPV4_PATTERN_BASE + u'$'
CIDRV4_PATTERN = EPSILON_PATTERN + '|^(' + IPV4_PATTERN_BASE + \
    u'(/(1[0-2][0-8]|[1-9]?[0-9]))' + u')$'

IPV6_PATTERN_BASE = (u'('
                     u'([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|'
                     u'([0-9a-fA-F]{1,4}:){1,7}:|'
                     u'([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|'
                     u'([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|'
                     u'([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|'
                     u'([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|'
                     u'([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|'
                     u'[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|'
                     u':((:[0-9a-fA-F]{1,4}){1,7}|:)|'
                     u'fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|'
                     u'::(ffff(:0{1,4}){0,1}:){0,1}'
                     u'((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}'
                     u'(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|'
                     u'([0-9a-fA-F]{1,4}:){1,4}:'
                     u'((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\\.){3,3}'
                     u'(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))')
IPV6_PATTERN = EPSILON_PATTERN + u'|^' + IPV6_PATTERN_BASE + u'$'
CIDRV6_PATTERN = EPSILON_PATTERN + u'|^(' + IPV6_PATTERN_BASE + \
    u'(/(1[0-2][0-8]|[1-9]?[0-9]))' + u')$'
IPV4_OR_IPV6_PATTERN = IPV4_PATTERN + u'|^' + IPV6_PATTERN_BASE + u'$'
CIDRV4_OR_CIDRV6_PATTERN = CIDRV4_PATTERN + u'|^' + CIDRV6_PATTERN + u'$'
UUID_BASE = u'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
UUID_PATTERN = EPSILON_PATTERN + u'|' + UUID_BASE

COMMONS = {
    u'description': u'Common data schemata shared among other schemata.',
    u'links': [],
    u'title': u'Kuryr Common Data Schema Definitions',
    u'properties': {
        u'options': {u'$ref': u'/schemata/commons#/definitions/options'},
        u'mac': {u'$ref': u'/schemata/commons#/definitions/mac'},
        u'cidrv6': {u'$ref': u'/schemata/commons#/definitions/cidrv6'},
        u'interface': {u'$ref': u'/schemata/commons#/definitions/interface'},
        u'cidr': {u'$ref': u'/schemata/commons#/definitions/cidr'},
        u'id': {u'$ref': u'/schemata/commons#/definitions/id'},
        u'uuid': {u'$ref': u'/schemata/commons#/definitions/uuid'},
        u'ipv4': {u'$ref': u'/schemata/commons#/definitions/ipv4'},
        u'ipv4_or_ipv6': {
            u'$ref': u'/schemata/commons#/definitions/ipv4_or_ipv6'}
    },
    u'definitions': {
        u'options': {
            u'type': [u'object', u'null'],
            u'description': u'Options.',
            u'example': {}
        },
        u'mac': {
            u'pattern': (EPSILON_PATTERN + u'|'
                         u'^((?:[0-9a-f]{2}:){5}[0-9a-f]{2}|'
                         u'(?:[0-9A-F]{2}:){5}[0-9A-F]{2})$'),
            u'type': u'string',
            u'description': u'A MAC address.',
            u'example': u'aa:bb:cc:dd:ee:ff'
        },
        u'cidrv6': {
            u'pattern': CIDRV6_PATTERN,
            u'type': u'string',
            u'description': u'A IPv6 CIDR of the subnet'
        },
        u'interface': {
            u'properties': {
                u'ID': {
                    u'description': u'Index of the interface',
                    u'type': u'number',
                },
                u'AddressIPv6': {
                    u'description': u'IPv6 CIDR',
                    u'$ref': u'#/definitions/commons/definitions/cidrv6'
                },
                u'MacAddress': {
                    u'description': u'MAC address',
                    u'$ref': u'#/definitions/commons/definitions/mac'
                },
                u'Address': {
                    u'description': u'IPv4 CIDR',
                    u'$ref': u'#/definitions/commons/definitions/cidr'
                }
            },
            u'type': [u'object', u'null'],
            u'description': u'Interface used in requests against Endpoints.',
            u'example': {
                u'AddressIPv6': u'fe80::f816:3eff:fe20:57c3/64',
                u'MacAddress': u'fa:16:3e:20:57:c3',
                u'Address': u'192.168.1.42/24'
            }
        },
        u'cidr': {
            u'pattern': CIDRV4_PATTERN,
            u'type': u'string',
            u'description': u'A IPv4 CIDR of the subnet.',
            u'example': u'10.0.0.0/24'
        },
        u'cidrv4_or_cidrv6': {
            u'pattern': CIDRV4_OR_CIDRV6_PATTERN,
            u'type': u'string',
            u'description': u'An IPv4 or IPv6 CIDR of the subnet.'
        },
        u'id': {
            u'oneOf': [
                {u'pattern': u'^([0-9a-f]{64})$'},
                {u'pattern': u'^([0-9a-z]{25})$'}],
            u'type': u'string',
            u'description': u'64 or 25 length ID value of Docker.',
            u'example': [
                u'51c75a2515d47edecc3f720bb541e287224416fb66715eb7802011d6ffd4'
                u'99f1',
                u'xqqzd9p112o4kvok38n3caxjm'
            ]
        },
        u'ipv4': {
            u'pattern': IPV4_PATTERN,
            u'type': u'string',
            u'description': u'An IPv4 address',
            u'example': u'10.0.0.1'
        },
        u'ipv4datum': {
            u'description': u'IPv4 data',
            u'required': [
                u'AddressSpace', u'Pool'],
            u'type': u'object',
            u'example': {
                u'AddressSpace': u'foo',
                u'Pool': u'192.168.42.0/24',
                u'Gateway': u'192.168.42.1/24',
                u'AuxAddresses': {
                    u'web': u'192.168.42.2',
                    u'db': u'192.168.42.3'
                }
            },
            u'properties': {
                u'AddressSpace': {
                    u'description': u'The name of the address space.',
                    u'type': u'string',
                    u'example': u'foo',
                },
                u'Pool': {
                    u'description': u'A range of IP Addresses requested in '
                                    u'CIDR format address/mask.',
                    u'$ref': u'#/definitions/commons/definitions/cidr'
                },
                u'Gateway': {
                    u'description': u'Optionally, the IPAM driver may provide '
                                    u'a Gateway for the subnet represented by '
                                    u'the Pool.',
                    u'$ref': u'#/definitions/commons/definitions/cidr',
                },
                u'AuxAddresses': {
                    u'description': u'A list of pre-allocated ip-addresses '
                                    u'with an associated identifier as '
                                    u'provided by the user to assist network '
                                    u'driver if it requires specific '
                                    u'ip-addresses for its operation.',
                    u'type': u'object',
                    u'patternProperties': {
                        u'.+': {
                            u'description': u'key-value pair of the ID and '
                                            u'the IP address',
                            u'$ref': u'#/definitions/commons/definitions/ipv4'
                        }
                    }
                }
            }
        },
        u'ipv6': {
            u'pattern': IPV6_PATTERN,
            u'type': u'string',
            u'description': u'An IPv6 address.',
            u'example': u'fe80::f816:3eff:fe20:57c4'
        },
        u'ipv6datum': {
            u'description': u'IPv6 data',
            u'required': [
                u'AddressSpace', u'Pool', u'Gateway'],
            u'type': u'object',
            u'example': {
                u'AddressCpace': u'bar',
                u'Pool': u'fe80::/64',
                u'Gateway': u'fe80::f816:3eff:fe20:57c3/64',
                u'AuxAddresses': {
                    u'web': u'fe80::f816:3eff:fe20:57c4',
                    u'db': u'fe80::f816:3eff:fe20:57c5'
                }
            },
            u'properties': {
                u'AddressSpace': {
                    u'description': u'The name of the address space.',
                    u'type': u'string',
                    u'example': u'foo',
                },
                u'Pool': {
                    u'description': u'A range of IP Addresses requested in '
                                    u'CIDR format address/mask.',
                    u'$ref': u'#/definitions/commons/definitions/cidrv6'
                },
                u'Gateway': {
                    u'description': u'Optionally, the IPAM driver may provide '
                                    u'a Gateway for the subnet represented by '
                                    u'the Pool.',
                    u'$ref': u'#/definitions/commons/definitions/cidrv6',
                },
                u'AuxAddresses': {
                    u'description': u'A list of pre-allocated ip-addresses '
                                    u'with an associated identifier as '
                                    u'provided by the user to assist network '
                                    u'driver if it requires specific '
                                    u'ip-addresses for its operation.',
                    u'type': u'object',
                    u'patternProperties': {
                        u'.+': {
                            u'description': u'key-vavule pair of the ID and '
                                            u'the IP address',
                            u'$ref': u'#/definitions/commons/definitions/ipv6'
                        }
                    }
                }
            }
        },
        u'ipv4_or_ipv6': {
            u'pattern': IPV4_OR_IPV6_PATTERN,
            u'type': u'string',
            u'description': u'An IPv4 or IPv6 address.',
            u'example': u'fe80::f816:3eff:fe20:57c4'
        },
        u'sandbox_key': {
            u'pattern': u'^(/var/run/docker/netns/[0-9a-f]{12})$',
            u'type': u'string',
            u'description': u'Sandbox information of netns.',
            u'example': '/var/run/docker/netns/12bbda391ed0'
        },
        u'uuid': {
            u'pattern': UUID_PATTERN,
            u'type': u'string',
            u'description': u'uuid of neutron resources.',
            u'example': 'dfe39822-ad5e-40bd-babd-3954113b3687'
        }
    },
    u'$schema': u'http://json-schema.org/draft-04/hyper-schema',
    u'type': u'object',
    u'id': u'schemata/commons'
}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt

from oslo_utils import importutils

from kuryr.lib import exceptions
from kuryr_libnetwork.port_driver import driver
from kuryr_libnetwork.tests.unit import base


@ddt.ddt
class TestDriver(base.TestKuryrBase):
    """Unit tests for driver loading"""

    @mock.patch.object(driver, '_verify_binding_driver_compatibility')
    @mock.patch.object(driver, '_verify_port_driver_compliancy')
    @mock.patch.object(importutils, 'import_object')
    @mock.patch.object(driver, '_parse_port_driver_config')
    def test_get_driver_instance(
            self, mock_parse_config, mock_import_object,
            mock_verify_compliancy, mock_verify_compatibility):
        module = 'kuryr_libnetwork.port_driver.drivers.veth'
        mock_parse_config.return_value = (module, 'veth', 'VethDriver')
        fake_driver = mock.Mock(spec=driver.Driver)
        mock_import_object.return_value = fake_driver

        response_driver = driver.get_driver_instance()
        mock_parse_config.assert_called_once()
        mock_import_object.assert_called_once_with(module + '.VethDriver')
        mock_verify_compliancy.assert_called_once_with(fake_driver, 'veth')
        mock_verify_compatibility.assert_called_once_with(fake_driver, 'veth')
        self.assertEqual(response_driver, fake_driver)

    @mock.patch('kuryr_libnetwork.config.CONF')
    @ddt.data('kuryr_libnetwork.port_driver.drivers.veth', 'veth')
    def test__parse_port_driver_config(self, port_driver_value, mock_conf):
        mock_conf.default_port_driver = port_driver_value

        module, name, classname = driver._parse_port_driver_config()
        self.assertEqual(module, 'kuryr_libnetwork.port_driver.drivers.veth')
        self.assertEqual(name, 'veth')
        self.assertEqual(classname, 'VethDriver')

    def test__verify_port_driver_compliancy(self):
        fake_driver = mock.Mock(spec=driver.Driver)
        ret = driver._verify_port_driver_compliancy(fake_driver, 'driver')
        self.assertIsNone(ret)

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__verify_binding_driver_compatibility(self, mock_conf):
        mock_conf.binding.enabled_drivers = ['veth']
        fake_driver = mock.Mock(spec=driver.Driver)
        fake_driver.get_supported_bindings.return_value = ('veth',)

        ret = driver._verify_binding_driver_compatibility(fake_driver, 'veth')
        fake_driver.get_supported_bindings.assert_called_once()
        self.assertIsNone(ret)

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__verify_binding_driver_compatibility_multi_drivers(
            self, mock_conf):
        mock_conf.binding.enabled_drivers = ['veth', 'sriov']
        fake_driver = mock.Mock(spec=driver.Driver)
        fake_driver.get_supported_bindings.return_value = ('sriov',)

        ret = driver._verify_binding_driver_compatibility(fake_driver, 'sriov')
        fake_driver.get_supported_bindings.assert_called_once()
        self.assertIsNone(ret)


class TestNestedDriverFailures(base.TestKuryrFailures):
    """Unit tests for driver loading failures"""

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__parse_port_driver_config_empty(self, mock_conf):
        mock_conf.default_port_driver = ''

        self.assertRaisesRegex(exceptions.KuryrException,
            "No port driver provided", driver._parse_port_driver_config)

    @mock.patch.object(importutils, 'import_object', side_effect=ImportError)
    def test_get_driver_instance_import_error(self, mock_import_object):
        self.assertRaises(exceptions.KuryrException,
            driver.get_driver_instance)

    def test__verify_port_driver_compliancy(self):
        class InvalidDriver(object):
            pass

        self.assertRaises(exceptions.KuryrException,
            driver._verify_port_driver_compliancy, InvalidDriver(), 'invalid')

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__verify_binding_driver_compatibility_not_compatible(self, m_conf):
        m_conf.binding.enabled_drivers = ['macvlan']
        message = r"Configuration file error: port driver 'veth' is not " \
                  r"compatible with binding driver '\['macvlan'\]'"

        fake_driver = mock.Mock(spec=driver.Driver)
        fake_driver.get_supported_bindings.return_value = ('veth',)
        self.assertRaisesRegex(exceptions.KuryrException, message,
            driver._verify_binding_driver_compatibility, fake_driver, 'veth')

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__verify_binding_driver_compatibility_not_compatible_multi_drivers(
            self, m_conf):
        m_conf.binding.enabled_drivers = ['macvlan', 'sriov']
        message = r"Configuration file error: port driver 'veth' is not " \
                  r"compatible with binding driver '\['macvlan'\, 'sriov']'"

        fake_driver = mock.Mock(spec=driver.Driver)
        fake_driver.get_supported_bindings.return_value = ('veth',)
        self.assertRaisesRegex(exceptions.KuryrException, message,
            driver._verify_binding_driver_compatibility, fake_driver, 'veth')

    @mock.patch('kuryr_libnetwork.config.CONF')
    def test__verify_binding_driver_compatibility_not_supported(self, m_conf):
        m_conf.binding.enabled_drivers = ['ipvlan']
        message = r"Configuration file error: binding driver " \
                  r"'\['ipvlan'\]' is currently not supported " \
                  r"with 'nested' port driver"

        fake_driver = mock.Mock(spec=driver.Driver)
        fake_driver.get_supported_bindings.return_value = ('ipvlan',)
        self.assertRaisesRegex(exceptions.KuryrException, message,
            driver._verify_binding_driver_compatibility, fake_driver, 'nested')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ddt
from oslo_serialization import jsonutils
from werkzeug import exceptions as w_exceptions

from kuryr.lib import utils as lib_utils
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils


@ddt.ddt
class TestKuryrJoinFailures(base.TestKuryrFailures):
    """Unit tests for the failures for binding a Neutron port."""
    def _invoke_join_request(self, docker_network_id,
                             docker_endpoint_id, container_id):
        data = {
            'NetworkID': docker_network_id,
            'EndpointID': docker_endpoint_id,
            'SandboxKey': utils.get_sandbox_key(container_id),
            'Options': {},
        }
        response = self.app.post('/NetworkDriver.Join',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))

        return response

    def test_join_bad_request(self):
        fake_docker_network_id = lib_utils.get_hash()
        invalid_docker_endpoint_id = 'id-should-be-hexdigits'
        fake_container_id = lib_utils.get_hash()

        response = self._invoke_join_request(
            fake_docker_network_id, invalid_docker_endpoint_id,
            fake_container_id)

        self.assertEqual(
            w_exceptions.BadRequest.code, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        # TODO(tfukushima): Add the better error message validation.
        self.assertIn(invalid_docker_endpoint_id, decoded_json['Err'])
        self.assertIn('Failed validating ', decoded_json['Err'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ddt import data
from ddt import ddt
from neutronclient.common import exceptions
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from kuryr.lib import utils as lib_utils
from kuryr_libnetwork import constants as const
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils


class TestKuryrNetworkCreateFailures(base.TestKuryrFailures):
    """Unittests for the failures for creating networks.

    This test covers error responses listed in the spec:
      http://developer.openstack.org/api-ref-networking-v2-ext.html#createProviderNetwork  # noqa
    """
    def _invoke_create_request(self, network_request):
        response = self.app.post('/NetworkDriver.CreateNetwork',
                                 content_type='application/json',
                                 data=jsonutils.dumps(network_request))
        return response

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_network')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.driver.get_default_network_id',
        return_value=None)
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_create_network_unauthorized(self, mock_list_subnetpools,
            mock_get_default_network, mock_create_network):
        docker_network_id = lib_utils.get_hash()
        network_request = {
            'NetworkID': docker_network_id,
            'IPv4Data': [{
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {}
            }],
            'IPv6Data': [{
                'AddressSpace': 'bar',
                'Pool': 'fe80::/64',
                'Gateway': 'fe80::f816:3eff:fe20:57c3/64',
                'AuxAddresses': {}
            }],
            'Options': {}
        }

        fake_kuryr_v4_subnetpool_id = uuidutils.generate_uuid()
        fake_v4_pool_name = lib_utils.get_neutron_subnetpool_name(
            network_request['IPv4Data'][0]['Pool'])
        kuryr_v4_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_v4_subnetpool_id, name=fake_v4_pool_name)

        fake_kuryr_v6_subnetpool_id = uuidutils.generate_uuid()
        fake_v6_pool_name = lib_utils.get_neutron_subnetpool_name(
            network_request['IPv6Data'][0]['Pool'])
        kuryr_v6_subnetpools = self._get_fake_v6_subnetpools(
            fake_kuryr_v6_subnetpool_id, name=fake_v6_pool_name)
        fake_v4_pool_attrs = {'name': fake_v4_pool_name}
        fake_v6_pool_attrs = {'name': fake_v6_pool_name}
        mock_list_subnetpools.side_effect = [
            {'subnetpools': kuryr_v4_subnetpools['subnetpools']},
            {'subnetpools': kuryr_v6_subnetpools['subnetpools']}
        ]

        fake_request = {
            "network": {
                "name": utils.make_net_name(docker_network_id),
                "admin_state_up": True,
                "shared": False
            }
        }
        mock_create_network.side_effect = exceptions.Unauthorized
        response = self._invoke_create_request(network_request)
        self.assertEqual(401, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        mock_list_subnetpools.assert_any_call(**fake_v4_pool_attrs)
        mock_list_subnetpools.assert_any_call(**fake_v6_pool_attrs)
        mock_create_network.assert_called_with(fake_request)
        self.assertIn('Err', decoded_json)
        self.assertEqual(
            {'Err': exceptions.Unauthorized.message}, decoded_json)

    @mock.patch(
        'kuryr_libnetwork.controllers.app.driver.get_default_network_id')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_create_network_get_default_network_id_unauthorized(self,
            mock_list_subnetpools, mock_get_default_network_id):
        docker_network_id = lib_utils.get_hash()
        network_request = {
            'NetworkID': docker_network_id,
            'IPv4Data': [{
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {}
            }],
            'IPv6Data': [{
                'AddressSpace': 'bar',
                'Pool': 'fe80::/64',
                'Gateway': 'fe80::f816:3eff:fe20:57c3/64',
                'AuxAddresses': {}
            }],
            'Options': {}
        }

        fake_kuryr_v4_subnetpool_id = uuidutils.generate_uuid()
        fake_v4_pool_name = lib_utils.get_neutron_subnetpool_name(
            network_request['IPv4Data'][0]['Pool'])
        kuryr_v4_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_v4_subnetpool_id, name=fake_v4_pool_name)

        fake_kuryr_v6_subnetpool_id = uuidutils.generate_uuid()
        fake_v6_pool_name = lib_utils.get_neutron_subnetpool_name(
            network_request['IPv6Data'][0]['Pool'])
        kuryr_v6_subnetpools = self._get_fake_v6_subnetpools(
            fake_kuryr_v6_subnetpool_id, name=fake_v6_pool_name)
        fake_v4_pool_attrs = {'name': fake_v4_pool_name}
        fake_v6_pool_attrs = {'name': fake_v6_pool_name}
        mock_list_subnetpools.side_effect = [
            {'subnetpools': kuryr_v4_subnetpools['subnetpools']},
            {'subnetpools': kuryr_v6_subnetpools['subnetpools']}
        ]

        mock_get_default_network_id.side_effect = exceptions.Unauthorized
        response = self._invoke_create_request(network_request)
        self.assertEqual(401, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        mock_list_subnetpools.assert_any_call(**fake_v4_pool_attrs)
        mock_list_subnetpools.assert_any_call(**fake_v6_pool_attrs)
        mock_get_default_network_id.assert_called()
        self.assertIn('Err', decoded_json)
        self.assertEqual(
            {'Err': exceptions.Unauthorized.message}, decoded_json)

    def test_create_network_bad_request(self):
        invalid_docker_network_id = 'id-should-be-hexdigits'
        network_request = {
            'NetworkID': invalid_docker_network_id,
            'IPv4Data': [{
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {}
            }],
            'IPv6Data': [{
                'AddressSpace': 'bar',
                'Pool': 'fe80::/64',
                'Gateway': 'fe80::f816:3eff:fe20:57c3/64',
                'AuxAddresses': {}
            }],
            'Options': {}
        }

        response = self._invoke_create_request(network_request)
        self.assertEqual(400, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        # TODO(tfukushima): Add the better error message validation.
        self.assertIn(invalid_docker_network_id, decoded_json['Err'])
        self.assertIn('Failed validating ', decoded_json['Err'])


@ddt
class TestKuryrNetworkDeleteFailures(base.TestKuryrFailures):
    """Unittests for the failures for deleting networks.

    This test covers error responses listed in the spec:
      http://developer.openstack.org/api-ref-networking-v2-ext.html#deleteProviderNetwork  # noqa
    """
    def _invoke_delete_request(self, network_name):
        data = {'NetworkID': network_name}
        response = self.app.post('/NetworkDriver.DeleteNetwork',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))
        return response

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_network')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_subnet')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_networks')
    @data(exceptions.Unauthorized, exceptions.NotFound, exceptions.Conflict)
    def test_delete_network_failures(self, GivenException,
            mock_list_networks, mock_list_subnets,
            mock_list_subnetpools, mock_delete_subnet,
            mock_delete_network):
        fake_subnetpools_response = {"subnetpools": []}
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        fake_neutron_network_id = "4e8e5957-649f-477b-9e5b-f1f75b21c03c"
        t = utils.make_net_tags(docker_network_id)
        te = t + ',' + utils.existing_net_tag(docker_network_id)
        subnet_v4_id = "9436e561-47bf-436a-b1f1-fe23a926e031"
        subnet_v6_id = "64dd4a98-3d7a-4bfd-acf4-91137a8d2f51"

        def mock_network(*args, **kwargs):
            if kwargs['tags'] == te:
                return self._get_fake_list_network(
                    fake_neutron_network_id,
                    check_existing=True)
            elif kwargs['tags'] == t:
                if GivenException == exceptions.NotFound:
                    return self._get_fake_list_network(
                        fake_neutron_network_id,
                        check_existing=True)
                return self._get_fake_list_network(
                    fake_neutron_network_id)

        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id)
        fake_v6_subnet = self._get_fake_v6_subnet(
            docker_network_id, docker_endpoint_id, subnet_v6_id)
        fake_subnets_response = {
            "subnets": [
                fake_v4_subnet['subnet'],
                fake_v6_subnet['subnet']
            ]
        }
        mock_list_networks.side_effect = mock_network
        mock_list_subnets.return_value = fake_subnets_response
        mock_list_subnetpools.return_value = fake_subnetpools_response
        mock_delete_subnet.return_value = None
        mock_delete_network.side_effect = GivenException

        response = self._invoke_delete_request(docker_network_id)
        decoded_json = jsonutils.loads(response.data)
        if GivenException == exceptions.NotFound:
            self.assertEqual(GivenException.status_code, 404)
            self.assertEqual(const.SCHEMA['SUCCESS'], decoded_json)
        else:
            self.assertEqual(GivenException.status_code, response.status_code)
            fake_v4_pool_attrs = {'name': 'kuryr'}
            fake_v6_pool_attrs = {'name': 'kuryr6'}
            mock_list_subnetpools.assert_any_call(**fake_v6_pool_attrs)
            mock_list_subnetpools.assert_any_call(**fake_v4_pool_attrs)
            mock_delete_subnet.assert_any_call(subnet_v4_id)
            mock_delete_subnet.assert_any_call(subnet_v6_id)
            mock_delete_network.assert_called_with(fake_neutron_network_id)
            mock_list_subnets.assert_called_with(
                network_id=fake_neutron_network_id)
            self.assertIn('Err', decoded_json)
            self.assertEqual({'Err': GivenException.message}, decoded_json)
        mock_list_networks.assert_any_call(tags=t)
        mock_list_networks.assert_any_call(tags=te)

    def test_delete_network_bad_request(self):
        invalid_docker_network_id = 'invalid-network-id'

        response = self._invoke_delete_request(invalid_docker_network_id)

        self.assertEqual(400, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        self.assertIn(invalid_docker_network_id, decoded_json['Err'])
        self.assertIn('Failed validating ', decoded_json['Err'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_subnet')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_networks')
    @data(exceptions.Unauthorized, exceptions.NotFound, exceptions.Conflict)
    def test_delete_network_with_subnet_deletion_failures(self,
            GivenException, mock_list_networks, mock_list_subnets,
            mock_list_subnetpools, mock_delete_subnet):

        fake_subnetpools_response = {"subnetpools": []}
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        fake_neutron_network_id = "4e8e5957-649f-477b-9e5b-f1f75b21c03c"
        t = utils.make_net_tags(docker_network_id)
        te = t + ',' + utils.existing_net_tag(docker_network_id)
        subnet_v4_id = "9436e561-47bf-436a-b1f1-fe23a926e031"
        subnet_v6_id = "64dd4a98-3d7a-4bfd-acf4-91137a8d2f51"

        def mock_network(*args, **kwargs):
            if kwargs['tags'] == te:
                return self._get_fake_list_network(
                    fake_neutron_network_id,
                    check_existing=True)
            elif kwargs['tags'] == t:
                return self._get_fake_list_network(
                    fake_neutron_network_id)

        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id)
        fake_v6_subnet = self._get_fake_v6_subnet(
            docker_network_id, docker_endpoint_id, subnet_v6_id)
        fake_subnets_response = {
            "subnets": [
                fake_v4_subnet['subnet'],
                fake_v6_subnet['subnet']
            ]
        }
        mock_list_networks.side_effect = mock_network
        mock_list_subnets.return_value = fake_subnets_response
        mock_list_subnetpools.return_value = fake_subnetpools_response
        mock_delete_subnet.side_effect = GivenException
        response = self._invoke_delete_request(docker_network_id)

        self.assertEqual(GivenException.status_code, response.status_code)
        mock_list_networks.assert_any_call(tags=t)
        mock_list_networks.assert_any_call(tags=te)
        mock_list_subnets.assert_called_with(
            network_id=fake_neutron_network_id)
        fake_v4_pool_attrs = {'name': 'kuryr'}
        fake_v6_pool_attrs = {'name': 'kuryr6'}
        mock_list_subnetpools.assert_any_call(**fake_v6_pool_attrs)
        mock_list_subnetpools.assert_any_call(**fake_v4_pool_attrs)
        mock_delete_subnet.assert_called_with(subnet_v4_id)

        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        self.assertEqual({'Err': GivenException.message}, decoded_json)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_utils import uuidutils

from kuryr.lib import binding
from kuryr.lib.binding.drivers import utils
from kuryr.lib import constants as lib_const
from kuryr.lib import exceptions
from kuryr.lib import segmentation_type_drivers as seg_driver
from kuryr.lib import utils as lib_utils
from kuryr_libnetwork.port_driver.drivers import vlan
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils as libnet_utils

mock_interface = mock.MagicMock()


class TestVlanDriver(base.TestKuryrBase):
    """Unit tests for the VlanDriver port driver"""

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    def test_get_supported_bindings(self, mock_trunk_port, mock_vlan_check):
        mock_trunk_port.return_value = None
        mock_vlan_check.return_value = None
        vlan_driver = vlan.VlanDriver()
        bindings = vlan_driver.get_supported_bindings()
        self.assertEqual(bindings, vlan.VlanDriver.BINDING_DRIVERS)

    @mock.patch('kuryr_libnetwork.config.CONF')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch.object(binding, 'port_bind')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_segmentation_id')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    def test_create_host_iface(self, mock_get_port_from_host,
                               mock_segmentation_id,
                               mock_port_bind, mock_vlan_check, mock_conf):
        mock_vlan_check.return_value = None
        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vm_port_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.3', 'fe80::f816:3eff:fe1c:36a9')['port']
        fake_neutron_port['mac_address'] = 'fa:16:3e:20:57:c3'
        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_vm_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.2', 'fe80::f816:3eff:fe20:57c4')['port']
        fake_vm_port['allowed_address_pairs'] = [
            {'ip_address': '192.168.1.2',
             'mac_address': fake_vm_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe20:57c4',
             'mac_address': fake_vm_port['mac_address']}]

        fake_subnets = self._get_fake_subnets(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)['subnets']

        fake_network = mock.sentinel.binding_network
        mock_conf.binding.link_iface = 'eth0'
        fake_exec_response = ('fake_stdout', '')
        fake_segmentation_id = 1
        mock_port_bind.return_value = ('fake_host_ifname',
            'fake_container_ifname', fake_exec_response)
        mock_segmentation_id.return_value = fake_segmentation_id
        mock_get_port_from_host.return_value = fake_vm_port

        vlan_driver = vlan.VlanDriver()

        response = vlan_driver.create_host_iface(fake_endpoint_id,
            fake_neutron_port, fake_subnets, fake_network)

        mock_get_port_from_host.assert_called_with(
            mock_conf.binding.link_iface)
        mock_port_bind.assert_called_with(fake_endpoint_id,
            fake_neutron_port, fake_subnets, fake_network, fake_vm_port,
            fake_segmentation_id)
        mock_segmentation_id.assert_called_with(fake_neutron_port['id'])

        self.assertEqual(response, fake_exec_response)

    @mock.patch('kuryr_libnetwork.config.CONF')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._release_segmentation_id')
    @mock.patch.object(binding, 'port_unbind')
    @mock.patch('kuryr_libnetwork.app.neutron.trunk_remove_subports')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    def test_delete_host_iface(self, mock_get_port_from_host,
                               mock_trunk_remove_subports, mock_port_unbind,
                               mock_release_seg_id, mock_vlan_check,
                               mock_conf):
        mock_vlan_check.return_value = None
        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_trunk_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vm_port_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id, fake_neutron_port_id,
            lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.3', 'fe80::f816:3eff:fe1c:36a9')['port']
        fake_neutron_port['mac_address'] = 'fa:16:3e:20:57:c3'
        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_vm_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.2', 'fe80::f816:3eff:fe20:57c4', 'fa:16:3e:20:57:c3',
            None, fake_neutron_trunk_id)['port']
        fake_vm_port['allowed_address_pairs'] = [
            {'ip_address': '192.168.1.3',
             'mac_address': fake_neutron_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe1c:36a9',
             'mac_address': fake_neutron_port['mac_address']}]

        mock_conf.binding.link_iface = 'eth0'
        fake_unbind_response = ('fake_stdout', '')
        mock_get_port_from_host.return_value = fake_vm_port
        mock_port_unbind.return_value = fake_unbind_response

        vlan_driver = vlan.VlanDriver()

        response = vlan_driver.delete_host_iface(fake_endpoint_id,
                                                 fake_neutron_port)

        mock_get_port_from_host.assert_called_with(
            mock_conf.binding.link_iface)
        mock_port_unbind.assert_called_with(fake_endpoint_id,
                                            fake_neutron_port)
        mock_trunk_remove_subports.assert_called_with(fake_neutron_trunk_id,
            {'sub_ports': [{
                'port_id': fake_neutron_port_id
            }]})
        mock_release_seg_id.assert_called_with(fake_neutron_port_id)

        self.assertEqual(response, fake_unbind_response)

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    @mock.patch.object(utils, 'get_veth_pair_names',
        return_value=("fake_host_ifname", "fake_container_name"))
    def test_get_container_iface_name(self, mock_get_pair_names,
                                      mock_trunk_port, mock_vlan_check):
        mock_trunk_port.return_value = None
        mock_vlan_check.return_value = None
        vlan_driver = vlan.VlanDriver()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_port = self._get_fake_port(
            uuidutils.generate_uuid(), uuidutils.generate_uuid(),
            fake_neutron_port_id)['port']
        response = vlan_driver.get_container_iface_name(fake_neutron_port)
        mock_get_pair_names.assert_called_with(fake_neutron_port_id)
        self.assertEqual(response, "fake_container_name")

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    @mock.patch('kuryr_libnetwork.app.neutron.trunk_add_subports')
    def test_attach_subport(self, mock_trunk_add_subports, mock_trunk_port,
                            mock_vlan_check):
        mock_trunk_port.return_value = None
        mock_vlan_check.return_value = None

        fake_neutron_trunk_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_segmentation_id = 1
        fake_subport = [
            {
                'segmentation_id': fake_segmentation_id,
                'port_id': fake_neutron_port_id,
                'segmentation_type': 'vlan'
            }
        ]

        vlan_driver = vlan.VlanDriver()

        vlan_driver._attach_subport(fake_neutron_trunk_id,
                                    fake_neutron_port_id,
                                    fake_segmentation_id)
        mock_trunk_add_subports.assert_called_with(fake_neutron_trunk_id,
                                                   {'sub_ports': fake_subport})

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    @mock.patch.object(seg_driver, 'allocate_segmentation_id')
    def test_get_segmentation_id(self, mock_alloc_seg_id, mock_trunk_port,
                                 mock_vlan_check):
        mock_trunk_port.return_value = None
        mock_vlan_check.return_value = None
        fake_neutron_port1_id = uuidutils.generate_uuid()
        fake_neutron_port2_id = uuidutils.generate_uuid()
        mock_alloc_seg_id.side_effect = [1, 2]

        vlan_driver = vlan.VlanDriver()

        response = vlan_driver._get_segmentation_id(fake_neutron_port1_id)
        mock_alloc_seg_id.assert_called_once()
        self.assertEqual(response, 1)

        mock_alloc_seg_id.reset_mock()
        response = vlan_driver._get_segmentation_id(fake_neutron_port1_id)
        mock_alloc_seg_id.assert_not_called()
        self.assertEqual(response, 1)

        response = vlan_driver._get_segmentation_id(fake_neutron_port2_id)
        mock_alloc_seg_id.assert_called_once()
        self.assertEqual(response, 2)

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._check_for_vlan_ids')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    @mock.patch('kuryr_libnetwork.app.neutron.update_port')
    @mock.patch.object(libnet_utils, 'get_neutron_port_name')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._attach_subport')
    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_segmentation_id')
    def test_update_port(self, mock_get_seg_id, mock_attach_subport,
                         mock_get_port_name, mock_update_port,
                         mock_get_port_from_host, mock_vlan_check):
        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_trunk_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vm_port_id = uuidutils.generate_uuid()

        fake_neutron_mac_address1 = 'fa:16:3e:20:57:c3'
        fake_neutron_mac_address2 = 'fa:16:3e:20:57:c4'
        fake_vm_mac_address = 'fa:16:3e:20:57:c5'
        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.3', 'fe80::f816:3eff:fe1c:36a9',
            fake_neutron_mac_address1,
            binding_host='', admin_state_up=False)['port']
        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_vm_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.2', 'fe80::f816:3eff:fe20:57c4', fake_vm_mac_address,
            None, fake_neutron_trunk_id)['port']
        fake_segmentation_id = 1
        fake_port_name = 'port1'

        mock_get_seg_id.return_value = fake_segmentation_id
        mock_get_port_name.return_value = fake_port_name
        mock_get_port_from_host.return_value = fake_vm_port
        mock_vlan_check.return_value = None

        vlan_driver = vlan.VlanDriver()

        vlan_driver.update_port(fake_neutron_port, fake_endpoint_id,
                                fake_neutron_mac_address2)

        mock_get_seg_id.assert_called_with(fake_neutron_port_id)

        mock_get_port_name.assert_called_with(fake_endpoint_id)

        mock_attach_subport.assert_called_with(fake_neutron_trunk_id,
                                               fake_neutron_port_id,
                                               fake_segmentation_id)

        mock_update_port.assert_called_with(fake_neutron_port_id,
                {'port': {
                    'device_owner': lib_const.DEVICE_OWNER,
                    'binding:host_id': lib_utils.get_hostname(),
                    'mac_address': fake_neutron_mac_address2,
                    'admin_state_up': True,
                }})


class TestVlanDriverFailures(base.TestKuryrFailures):
    """Unit tests for the VlanDriver port driver failures"""

    @mock.patch('kuryr_libnetwork.port_driver.drivers.vlan'
                '.VlanDriver._get_port_from_host_iface')
    def test_create_host_iface(self, mock_get_port_from_host):
        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE)['port']

        vlan_driver = vlan.VlanDriver()
        self.assertRaises(exceptions.KuryrException,
            vlan_driver.create_host_iface, fake_endpoint_id,
            fake_neutron_port, None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kuryr_libnetwork.schemata import endpoint_delete


LEAVE_SCHEMA = endpoint_delete.ENDPOINT_DELETE_SCHEMA
LEAVE_SCHEMA[u'title'] = u'Leave endpoint'
LEAVE_SCHEMA[u'links'] = [{
    u'method': u'POST',
    u'href': u'/NetworkDriver.Leave',
    u'description': u'Unbinds the endpoint from the container.',
    u'rel': u'self',
    u'title': u'Leave'
}]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import ddt

from kuryr.lib import utils as lib_utils
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils


@ddt.ddt
class TestKuryrUtils(base.TestKuryrBase):
    """Unit tests for utilities."""

    @ddt.data(lib_utils.get_hash(), '51c75a2515d4' '51c75a')
    def test_get_sandbox_key(self, fake_container_id):
        sandbox_key = utils.get_sandbox_key(fake_container_id)
        expected = '/'.join([lib_utils.DOCKER_NETNS_BASE,
                            fake_container_id[:12]])
        self.assertEqual(expected, sandbox_key)

    def test_get_port_name(self):
        fake_docker_endpoint_id = lib_utils.get_hash()
        generated_neutron_port_name = utils.get_neutron_port_name(
            fake_docker_endpoint_id)
        self.assertIn(lib_utils.PORT_POSTFIX, generated_neutron_port_name)
        self.assertIn(fake_docker_endpoint_id, generated_neutron_port_name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from kuryr.lib.binding.drivers import utils
from kuryr.lib import exceptions
from kuryr_libnetwork import app
from kuryr_libnetwork import config
from kuryr_libnetwork.port_driver import driver


class BaseNestedDriver(driver.Driver, metaclass=abc.ABCMeta):
    """Driver for container-in-VM deployments with MACVLAN and IPVLAN."""

    def __init__(self):
        self.link_iface = config.CONF.binding.link_iface

    def _get_port_from_host_iface(self, ifname):
        """Returns the Neutron port associated to ifname or raises otherwise.

        Returns the Neutron port associated to ifname if such port exists, a
        exceptions.KuryrException if it does not,
        n_exceptions.NeutronClientException on errors.

        :returns: a Neutron port dictionary as returned by
                  python-neutronclient or None
        :raises: exceptions.KuryrException
                 neutronclient.common.exceptions.NeutronClientException
        """
        ip = utils.get_ipdb()

        mac_address = ip.interfaces.get(ifname, {}).get('address', None)
        if mac_address:
            ports = app.neutron.list_ports(mac_address=mac_address)
            if ports['ports']:
                return ports['ports'][0]

        raise exceptions.KuryrException("Cannot find a Neutron port "
            "associated to interface name {0}".format(ifname))

    def get_container_iface_name(self, neutron_port):
        """Returns interface name of a container in the default namespace.

        :param neutron_port: The neutron port
        :returns: interface name as string.
        """
        _, container_iface_name = utils.get_veth_pair_names(neutron_port['id'])
        return container_iface_name
